//#####################################################################
//## AUTHOR:      Jaliya Ekanakayek (jekanaya@cs.indiana.edu)
//## FILE:        ClientRootUI.C
//##
//## DESCRIPTION: User interface to add Clarens servers and receive
//## 			  Notications when the files are been analysed.
//##
//## THANKS     : Source code published by Tayfun Ince (Tayfun.Ince@cern.ch)
//#####################################################################
#include <stdlib.h>
#include <string.h>
#include <map.h>
#include <vector.h>
#include <iostream>
#include <stdio.h>
#include <time.h>

#include <TROOT.h>
#include <TApplication.h>
#include <TGListBox.h>
#include <TGListTree.h>
#include <TGFSContainer.h>
#include <TGClient.h>
#include <TGFrame.h>
#include <TGIcon.h>
#include <TGLabel.h>
#include <TGButton.h>
#include <TGTextEntry.h>
#include <TGMsgBox.h>
#include <TGTextEdit.h>
#include <TGProgressBar.h>
#include <RQ_OBJECT.h>
#include <TH1.h>
#include <TH2.h>
#include <TRandom.h>
#include <TSystem.h>
#include <TSystemDirectory.h>
#include <TEnv.h>
#include <TFile.h>
#include <TGButton.h>
#include <TGDockableFrame.h>
#include <TRootEmbeddedCanvas.h>
#include <TGStatusBar.h>
#include <TGComboBox.h>
#include <TSystem.h>
#include <TH1D.h>
#include <TH1F.h>
#include <TString.h>
#include <TLegend.h>
#include <TPad.h>
#include <TCanvas.h>
#include <TImage.h>

#define Analyse_Eg05_cxx
#include "Analyse_Eg05.h"

Int_t FALSE = 0;
Int_t TRUE = 1;

#define NB_BROKER_HOST "156.56.104.95";
#define NB_BROKER_PORT 5045;
#define NB_ENTITY_ID_DIV_FACTOR 8000;
#define NB_TEMPLATE_ID 8899;
#define DEMO_TOPIC "/test/clarens/hep";
#define EVENT_TOPIC "/test/clarens/event";
#define SESSION_MGT_TOPIC "/test/session_mangement"
#define RESPONCE_TOPIC "/test/clarens/hep";

#define ANALYSIS_C_FILE 	"Analyse_Eg05.C";
#define ANALYSIS_H_FILE 	"Analyse_Eg05.h";

#define MAX_FILES_PER_ROOTLET 10;

#define ROOT_DATA 		"born_1.root";
#define OUTPUT_FILE_PREFIX  "histos.root";
#define OUTPUT_DATA "output.histo";
#define TMP_HISTO "tmp2.histo";

#define IMAGE_LOCATION "RootletArchitecture.jpg";


#define BEFORE_MERGE 1;
#define AFTER_MERGE 2;

#define SES_NAME_REQ "ses_name_req";
#define SES_NAME_RES "ses_name_res";
#define SES_DATA_REQ "ses_data_req";
#define SES_DATA_RES "ses_data_res"

#define MY_PI 3.141592654


using namespace std;

// Class header
class ClientRootUI : public TGMainFrame
{   
private:
    TGHorizontalFrame *contentFrame;
    TGHorizontalFrame *statusFrame;
    TGVertical3DLine *separator;

    TGVerticalFrame *controlFrame;
    TGHorizontalFrame *ctrBtnFrame;

    //To hold the available clarens servers.
    TGListBox *fListBox;
    TList *fSelected;

    TGHorizontalFrame *frmNumFiles;
    TGTextEntry *numFiles;
    TGLabel *lblNumFiles;
    TGTextButton *fLoad;
    TGTextButton *fAnalyse;

    //The canvas to add the labels for root files.
    TRootEmbeddedCanvas *rootEFileCanvas;
    TCanvas *rootFileCanvas;

    //All the servers initialized
    map<string, void*> clarensServers;
    //Servers of which the files are loaded for processing.
    int currentServerPosInCanvas;

    TRootEmbeddedCanvas *fECanvas;
    TGStatusBar *fStatusBar;

    TCanvas *fCanvas;
    TPad *pad;
    TPad *pad2;

    int m_canvasW, m_canvasH;
    void * tServiceClient;
    void * tCallbackData;
    void * tCallbackEvent;

    void *rFileCanvasMgr;
    
    string eventTopic;
    TImage *img;

public:
    ClientRootUI(const TGWindow *p, UInt_t w, UInt_t h);
    virtual ~ClientRootUI();
    void CloseMainWindow();

    void Connect();
    void Subscribe();

    void SetupContentFrame();
    void SetupControlFrame();
    void SetupOtherControls();

    void TryToClose();
    void AddClarensServer(string host);

    void ShowMessageBox(string type, string message);
    string getFileNameFromRemoteURL(string str);
    
    void callback_data(void **v, char *bytes, int length);


};

//#####################################################################
//#####################################################################
// Class constructor
ClientRootUI::ClientRootUI(const TGWindow *p, UInt_t w, UInt_t h)
{
    // a few gui notes:
    // do not share TGLayoutHints;
    // if using TGCompositeFrame (or other that inherits) as a container,
    // use TGCompositeFrame::SetCleanup(kDeepCleanup) in order to ensure
    // proper deletion of child gui object inside when destructer called;
    // when TGDockableFrame is used, need to delete all child objects
    // individually in the destructor.

    // initialize variables
    currentServerPosInCanvas=0;
    //m_canvasW = 720;
    //m_canvasH = 540;
    m_canvasW = 360;
    m_canvasH = 360;

    Connect("CloseWindow()", "ClientRootUI", this, "CloseMainWindow()");

    SetupContentFrame();
    SetupControlFrame();
    rFileCanvasMgr=new TFileCanvasManager(rootFileCanvas);
    SetupOtherControls();

    // set main window name and position
    SetWindowName("NBR-Clarens-ROOT Client");
    MapSubwindows();
    Resize(GetDefaultSize());

    // initialize layout of the main window and map the GUI
    Layout();
    MapWindow();

    //Random entity Id
    srand((unsigned) time(0));
    int entityID = (rand() % NB_ENTITY_ID_DIV_FACTOR) + NB_ENTITY_ID_DIV_FACTOR;
    
    stringstream ss;
    ss<<EVENT_TOPIC;
    ss<<"/";
    ss<<entityID;    
    this->eventTopic=ss.str();    

    //For the publish/subscribe messaging.
    tServiceClient=(TServiceClient *)(new TServiceClient());
    ((TServiceClient *)tServiceClient)->Init(NB_BROKER_HOST,NB_BROKER_PORT,entityID,NB_TEMPLATE_ID);

    tCallbackData=new TCallback();
    tCallbackEvent=new TCallback();

    fStatusBar-> ChangeBackground(46);
    fStatusBar->SetText("Connecting ...");
    gSystem->ProcessEvents();
    Subscribe();
    fStatusBar->SetText("Sucessfully Connected.");
    gSystem->ProcessEvents();
};

void ClientRootUI::SetupControlFrame()
{
    fListBox = new TGListBox(controlFrame, 100);
    fListBox->Selected(1);
    fListBox->Resize(300, 100);
    fListBox-> SetMultipleSelections(true);
    controlFrame->AddFrame(fListBox, new TGLayoutHints( kLHintsLeft |kLHintsExpandX, 5, 5, 5, 5));

    ctrBtnFrame= new TGHorizontalFrame(controlFrame, 200, 200);
    ctrBtnFrame->SetCleanup(kDeepCleanup);
    controlFrame->AddFrame(ctrBtnFrame, new TGLayoutHints(kLHintsCenterX,4,4,2,2));

    //Setup the controll frame
    fLoad = new TGTextButton(ctrBtnFrame," Connect To Session ");
    fLoad->Connect("Clicked()", "ClientRootUI", this, "Connect()");
    ctrBtnFrame->AddFrame(fLoad, new TGLayoutHints(kLHintsLeft,4,4,2,2));


    // add embedded canvas to the main window
    rootEFileCanvas = new TRootEmbeddedCanvas(0,controlFrame, 200,m_canvasH);
    rootEFileCanvas->SetAutoFit(kTRUE);
    controlFrame->AddFrame(rootEFileCanvas, new TGLayoutHints(kLHintsLeft| kLHintsExpandX | kLHintsExpandY));

    //Setup the control frame canvas.
    rootFileCanvas=rootEFileCanvas->GetCanvas();
    rootFileCanvas->SetGrid();
    rootFileCanvas->SetLeftMargin(0.15);
    rootFileCanvas->SetBottomMargin(0.15);
};

void ClientRootUI::SetupContentFrame()
{
    //Add a horizontal frame to hold the histogram and the locations sub frames
    contentFrame = new TGHorizontalFrame(this, 200, 200);
    contentFrame->SetCleanup(kDeepCleanup); // for proper deletion childs
    AddFrame(contentFrame, new TGLayoutHints(kLHintsCenterX| kLHintsExpandX | kLHintsExpandY,2,2,0,2));

    // add embedded canvas to the main window
    fECanvas = new TRootEmbeddedCanvas(0,contentFrame, m_canvasW, m_canvasH);
    contentFrame->AddFrame(fECanvas, new TGLayoutHints(kLHintsLeft | kLHintsExpandX | kLHintsExpandY));

    fCanvas = fECanvas->GetCanvas();
    fCanvas->SetDoubleBuffer(1);
    fCanvas->SetGrid();
    fCanvas->SetLeftMargin(0.15);
    fCanvas->SetBottomMargin(0.15);

    fCanvas->cd();

    pad = new TPad("pad","Pad",0.02,0.02,0.98,0.98);
    pad->Draw();

    pad->cd();
    img = TImage::Open(IMAGE_LOCATION);
    img->Draw();

    separator = new TGVertical3DLine(contentFrame);
    contentFrame->AddFrame(separator, new TGLayoutHints(kLHintsLeft | kLHintsExpandY,2,2,0,0));

    controlFrame=new TGVerticalFrame(contentFrame, 200, 200);
    controlFrame->SetCleanup(kDeepCleanup); // for proper deletion childs

    contentFrame->AddFrame(controlFrame, new TGLayoutHints(kLHintsLeft | kLHintsExpandY,2,2,0,2));
};

void ClientRootUI::TryToClose()
{
    // The user try to close the main window,
    //  while a message dialog box is still open.
    printf("Can't close the window '%s' : a message box is still open\n",     GetWindowName());
}
;

/*
 * Utility method to display a message box.
 * */
void ClientRootUI::ShowMessageBox(string type, string message)
{
   // cout<<"Showing message"<<endl;

    int i, buttons, retval;
    EMsgBoxIcon icontype = kMBIconStop;
    buttons = kMBDismiss;

    // Since the message dialog box is created, we disable the
    // signal/slot communication mechanism, in order to ensure we
    // can't close the this window while the message box is open.
    Disconnect("CloseMainWindow()");
    Connect("CloseWindow()", "ClientRootUI", this, "TryToClose()");
    //TODO: How to delete this pointer?
    new TGMsgBox(gClient->GetRoot(), this,type.c_str(), message.c_str(), icontype, buttons, &retval);
    Disconnect("CloseWindow()");
    Connect("CloseWindow()", "ClientRootUI", this, "CloseMainWindow()");

};

/*
 * Add a clarens object to available servers after connecting to it.
 * These clarens objects are already connected to a particular server
 * can directly use after verifying the connectivity.
 * */
void ClientRootUI::AddClarensServer(string host)
{
    TClarensServer *server=new TClarensServer(host);
    clarensServers.insert(make_pair(host, (void *)server ));
};

void ClientRootUI::SetupOtherControls()
{
    // add status bar for messages to the main window
    fStatusBar = new TGStatusBar(this, 50, 10, kHorizontalFrame);
    AddFrame(fStatusBar, new TGLayoutHints(kLHintsExpandX, 2, 2, 0, 2));
    fStatusBar-> SetHeight(30);
    fStatusBar->SetText("NBR-Clarens-ROOT UI.");
};

void ClientRootUI::Subscribe()
{
	pad->cd();
	
	 char tline[10000];

	   TLegend *legend = new TLegend(0.4,0.6,0.9,0.8);
	   legend->SetHeader("REMOTE Analysis");
	   //cerr << "Analysis using local .root file" << endl;

	   strcpy(tline, ".x ");
	   strcat(tline, ANALYSIS_C_FILE);
	   strcat(tline, "(\"");
	   strcat(tline, ROOT_DATA);
	   strcat(tline, "\",\"");
	   strcat(tline, OUTPUT_DATA);
	   strcat(tline, "\")");

	   //cerr << "Executing Local Analysis: " << tline << endl;
	   gROOT->ProcessLine(tline);

	   TFile *tflocal=new TFile(OUTPUT_DATA);
	   TH1F *hRecMass = (TH1F *) tflocal->Get("hRecMass1");
	   hRecMass->SetMarkerStyle(21);

	   gROOT->ProcessLine(".L fitMass.C");

    int count=0;
    void **args=new void *[3];
    args[0]=this;
    args[1]=legend;
    args[2]=hRecMass;

    /*Callback function to call when a message is received.*/
    ((TCallback *)tCallbackData)->registerCallback(args, "ClientRootUI", "callback_data");
    ((TServiceClient *)tServiceClient)->Subscribe(DEMO_TOPIC,tCallbackData);
    
     stringstream buff;
	 buff<<SES_NAME_REQ;
	 buff<<"|";
	 buff<<eventTopic;	
	 
	 int entityID = (rand() % NB_ENTITY_ID_DIV_FACTOR) + NB_ENTITY_ID_DIV_FACTOR;
	 TCall *tcall=new TCall();
	 tcall->Init(NB_BROKER_HOST,NB_BROKER_PORT,entityID,NB_TEMPLATE_ID);
	 char *response;
	 int resLen=0;
	 tcall->SendReceive(SESSION_MGT_TOPIC,eventTopic,(char*)(buff.str().c_str()),strlen(buff.str().c_str()),&response,&resLen);
	 printf("Response Length = %d \n",resLen);
	 string msg(response,resLen);
	    //First find the type of the message
	    int offset=0;
	    int index_one=msg.find("|");
	    if(index_one==string::npos)
	    {
	        return;
	    } //Not for me.
	    string msg_type=msg.substr(offset,index_one-offset);
	    offset+=index_one+1;
	    if(msg_type.compare(SES_NAME_RES)==0)
	    {
	    	string ses_name=msg.substr(offset);
	    	this->fListBox->NewEntry(ses_name.c_str());
	    	gSystem->ProcessEvents();  
	    
	    }    
};


/*
 * Simplify the UI functions
 * 
 * */
void ClientRootUI::Connect()
{
    fStatusBar-> ChangeBackground(46);
    fStatusBar->SetText("Connecting ...");
    gSystem->ProcessEvents();
	
	stringstream buff;
	buff<<SES_DATA_REQ;
	buff<<"|";
	buff<<this->eventTopic;
	
	 TFileCanvasManager *mgr=(TFileCanvasManager *)(this->rFileCanvasMgr);
	
	 int entityID = (rand() % NB_ENTITY_ID_DIV_FACTOR) + NB_ENTITY_ID_DIV_FACTOR;
	 TCall *tcall=new TCall();
	 tcall->Init(NB_BROKER_HOST,NB_BROKER_PORT,entityID,NB_TEMPLATE_ID);
	 char *response;
	 int resLen=0;
	 tcall->SendReceive(SESSION_MGT_TOPIC,eventTopic,(char*)(buff.str().c_str()),strlen(buff.str().c_str()),&response,&resLen);
	
	 string msg(response,resLen);
	     //First find the type of the message
	     int offset=0;
	     int index_one=msg.find("|");
	     if(index_one==string::npos)
	     {
	         return;
	     } //Not for me.
	     string msg_type=msg.substr(offset,index_one-offset);
	     offset+=index_one+1;
	 if(msg_type.compare(SES_DATA_RES)==0)
	     {
	         int num_servers=0;
	         int num_rootlets=0;

	         int index_two=msg.find("|",offset);
	         if(index_two==string::npos)
	         {
	             return;
	         }
	         string str_num_servers=msg.substr(offset,index_two-offset);
	         offset=index_two+1;
	         std::istringstream s1(str_num_servers);
	         s1 >> num_servers;

	         //cout<<"Offset Two ="<<offset<<endl;

	         int index_three=msg.find("|",offset);
	         if(index_three==string::npos)
	         {
	             return;
	         }
	         //cout<<"OOOOOOO  "<<offset<<"  xxx "<<index_three<<endl;
	         string str_num_rootlet=msg.substr(offset,index_three-offset);
	         offset=index_three+1;
	         std::istringstream s2(str_num_rootlet);
	         s2 >> num_rootlets;

	         bool end_of_str=false;
	         while(!end_of_str)
	         {
	             string str_server_name;
	             int index_x=msg.find("|",offset);

	          //   cout<<"Offset for severs = "<<offset<<endl;

	             if(index_x==string::npos)
	             {
	                // cout<<"Index npos servers = "<<index_x<<endl;
	                 end_of_str=true;
	                 str_server_name=msg.substr(offset);
	             }
	             else
	             {
	                // cout<<"Index in servers = "<<index_x<<endl;
	                 str_server_name=msg.substr(offset,index_x-offset);
	                 offset=index_x+1;
	             }
	             
	             int num_data_files=0;
	             int hash_pos=str_server_name.find("#");
	             string server_url=str_server_name.substr(0,hash_pos);
	             string str_num_files=str_server_name.substr(hash_pos+1);
	             std::istringstream ss(str_num_files);
	 	        ss >> num_data_files;
	 	     
	             this->AddClarensServer(server_url);

	             int endOfHTTPS=server_url.find(":", 0)+3;
	             int lengthOfFirstSegment=server_url.find(".", 0) -endOfHTTPS;
	             string lblStr=server_url.substr(endOfHTTPS, lengthOfFirstSegment);
	             mgr->addServer(server_url,lblStr,num_data_files);

	           //  cout<<"Added Server "<<server_url<<endl;
	         }
	         mgr->setNumFilePerRootlet(num_rootlets);
	         mgr->draw();
	         
	         this->fStatusBar->SetText("Successfully Connected to the Session");
	     	gSystem->ProcessEvents();
	         
	       
	     }
	     else
	     {
	         return;
	     }

	    // cout<<"Event Received :"<<msg<<endl;
	     delete[] response;    
};


//#####################################################################
// Class destructorl,
ClientRootUI::~ClientRootUI()
{

    Cleanup();
}
;

//#####################################################################
// Close main window upon click on "X"
void ClientRootUI::CloseMainWindow()
{
     ((TServiceClient *)tServiceClient)->Cleanup();
    delete this;
}
;

string ClientRootUI::getFileNameFromRemoteURL(string str)
{
    int lastIndex=0;
    bool found=false;
    string::size_type loc;
    for (int i=0; i<str.length(); i++)
    {
        loc = str.find("/", i);
        if (loc != string::npos)
        {
            //cout << "Found Omega at " << loc << endl;
            lastIndex=loc;
        }
        else
        {
            //cout << "Didn't find Omega" << endl;
            found =true;
        }
    }
    if (found)
    {
        return str.substr(lastIndex+1);
    }
    else
    {
        ShowMessageBox("Information", "Could not find the file name.");
        //TODO correct this.
        return "";
    }

}
;


void ClientRootUI::callback_data(void **v, char *bytes, int length)
{
	
	ClientRootUI *ui=(ClientRootUI *)v[0];
	TFileCanvasManager *mgr=(TFileCanvasManager *)(ui->rFileCanvasMgr);
	TLegend *remoteLegend=(TLegend *)v[1];
	TH1F *hRecMassRemote=(TH1F *)v[2];

    

    //Read the timestamp comming with the message///////////////////////////
    long long offset=0;
    long long time=TUtils::ReadLong(bytes,offset);

    const int header_len=TUtils::ReadInt(bytes+offset,offset);
    const int new_header_len=header_len;
    char header[new_header_len];
    TUtils::CopyBytes(bytes+offset,header,header_len,offset);
    string header_str(header,header_len);
    ////////////////////////////Update the Canvase/////////////////////////////////////////
    string::size_type loc = header_str.find("#", 0);
    if (loc == string::npos)
    {
        cerr<<"Could not understand the message "<<endl;
        return;
    }

    string serverUrl(header_str.c_str(), 0, loc);
    string shell_histo_name(header_str.c_str(), loc+1, length-loc-1);
    //cerr<<" Server Name = "<<serverUrl<<" Histogram File : "<<shell_histo_name   <<endl;

    string fileName=getFileNameFromRemoteURL(shell_histo_name);

    int num_pos_begin=fileName.find_first_of("_",0) +1; //Rootlet_x.histo;
    int num_pos_end=fileName.find_first_of(".",0); //Rootlet_x.histo;
    string str_num=fileName.substr(num_pos_begin,(num_pos_end-num_pos_begin));

    istringstream buffer(str_num);
    int rootlet_num;
    buffer >> rootlet_num;
    mgr->updateRootlet(serverUrl,rootlet_num, BEFORE_MERGE);

    /////////////////////////////End updating Canvas///////////////////////////////////////

    ui->fCanvas->cd();
    ui->pad->cd();
    
    ofstream myFile (TMP_HISTO, ios::trunc|ios::out | ios::binary);
    myFile.write (bytes+offset, length-offset);
    myFile.flush();
    myFile.close();                 
    
    //Check if the file exists else use the option NEW for the constructor.  
    TFile *tfile=new TFile(TMP_HISTO);    
    //tfile->WriteBuffer(bytes+offset,length-offset);   

    TH1F *hrm = (TH1F *) tfile->Get("hRecMass1");
   
    hRecMassRemote->Add(hrm);
    hRecMassRemote->SetMarkerStyle(21);
   
    fitMass(hRecMassRemote, remoteLegend);
    //fitMass(hrm, remoteLegend);

    remoteLegend->Draw();
    ui->fCanvas->Update();        
    
    mgr->updateRootlet(serverUrl,rootlet_num, AFTER_MERGE);
    
    delete tfile;
    

};

//########################### END OF CALL BACK FUNCTION #######################################


//#####################################################################
//#####################################################################
// Launch GUI
void ClientRootUI()
{
   gDebug = kFALSE;
   /**************************************************************************/
   gSystem->Load("libnbROOT.so"); //Added for nbclarens pubsub capabilities
   /**************************************************************************/

   gROOT->SetStyle("Plain");

    ClientRootUI *ui=new ClientRootUI(gClient->GetRoot(),200,200);
}
//#####################################################################


