#include <stdlib.h>

#include <TROOT.h>
#include <TApplication.h>
#include <TVirtualX.h>
#include <TVirtualPadEditor.h>
#include <TGResourcePool.h>
#include <TGListBox.h>
#include <TGListTree.h>
#include <TGFSContainer.h>
#include <TGClient.h>
#include <TGFrame.h>
#include <TGIcon.h>
#include <TGLabel.h>
#include <TGButton.h>
#include <TGTextEntry.h>
#include <TGNumberEntry.h>
#include <TGMsgBox.h>
#include <TGMenu.h>
#include <TGCanvas.h>
#include <TGComboBox.h>
#include <TGTab.h>
#include <TGSlider.h>
#include <TGDoubleSlider.h>
#include <TGFileDialog.h>
#include <TGTextEdit.h>
#include <TGShutter.h>
#include <TGProgressBar.h>
#include <TGColorSelect.h>
#include <RQ_OBJECT.h>
#include <TCanvas.h>
#include <TColor.h>
#include <TH1.h>
#include <TH2.h>
#include <TRandom.h>
#include <TSystem.h>
#include <TSystemDirectory.h>
#include <TEnv.h>
#include <TFile.h>
#include <TKey.h>
#include <TGDockableFrame.h>
#include <TGFontDialog.h>

#include <TGButton.h>
#include <TGDockableFrame.h>
#include <TRootEmbeddedCanvas.h>
#include <TGMenu.h>
#include <TPostScript.h>
#include <TCanvas.h>
#include <TStyle.h>
#include <TGStatusBar.h>
#include <TGComboBox.h>
#include <TGFileDialog.h>
#include <TGraph.h>
#include <TSystem.h>
#include <TH1D.h>
#include <TString.h>
#include <TImage.h>


#define Analyse_Eg05_cxx
#include "Analyse_Eg05.h"
#include <TLegend.h>
#include <TPad.h>
#include <TStyle.h>
#include <TCanvas.h>
#include <string.h>
#include <map.h>
#include <iostream>

#include <stdio.h>
#include <time.h>

Int_t FALSE = 0;
Int_t TRUE = 1;

#define NB_BROKER_HOST "156.56.104.95";
#define NB_BROKER_PORT 5045;
#define NB_ENTITY_ID_DIV_FACTOR 8000;
#define NB_TEMPLATE_ID 8899;
#define HISTO_VIEW_TOPIC "/test/shared/display";

using namespace std;

// Class header
class HistoViewer : public TGMainFrame
{

    //  RQ_OBJECT("RootUI") // necessary if not inherite from ROOT class

private:

    TGHorizontalFrame *contentFrame;

    TRootEmbeddedCanvas *fECanvas;
    TCanvas *fCanvas;
    TPad *pad;

    int m_canvasW, m_canvasH;

    void * tServiceClient;
    void * tCallback;
    TImage *img;

public:
	HistoViewer(const TGWindow *p, UInt_t w, UInt_t h);
    virtual ~HistoViewer();
    void Subscribe();
    void SetupContentFrame();
    void callback(void **v, char *bytes, int length);
    void TryToClose();
};

// Class constructor
HistoViewer::HistoViewer(const TGWindow *p, UInt_t w, UInt_t h)
{
    // a few gui notes:
    // do not share TGLayoutHints;
    // if using TGCompositeFrame (or other that inherits) as a container,
    // use TGCompositeFrame::SetCleanup(kDeepCleanup) in order to ensure
    // proper deletion of child gui object inside when destructer called;
    // when TGDockableFrame is used, need to delete all child objects
    // individually in the destructor.

    // initialize variables
    m_canvasW = 720;
    m_canvasH = 540;

    Connect("CloseWindow()", "RootUI", this, "CloseMainWindow()");

    //ShowMenu();
    SetupContentFrame();

    // set main window name and position
    SetWindowName("Histo Viewer");
    MapSubwindows();
    Resize(GetDefaultSize());
    //MoveResize(0,0,GetMaxHeight(),GetMaxWidth());

    // initialize layout of the main window and map the GUI
    Layout();
    MapWindow();

    //Random entity Id
    srand((unsigned) time(0));
    int entityID = (rand() % NB_ENTITY_ID_DIV_FACTOR) + NB_ENTITY_ID_DIV_FACTOR;

    //For the publish/subscribe messaging.
    tServiceClient=(TServiceClient *)(new TServiceClient());
    ((TServiceClient *)tServiceClient)->Init(NB_BROKER_HOST,NB_BROKER_PORT,entityID,NB_TEMPLATE_ID);
    tCallback=new TCallback();
    
  
};

// Class destructorl,
HistoViewer::~HistoViewer()
{
    Cleanup();
};


void HistoViewer::SetupContentFrame()
{
    //Add a horizontal frame to hold the histogram and the locations sub frames
    contentFrame = new TGHorizontalFrame(this, 200, 200);
    contentFrame->SetCleanup(kDeepCleanup); // for proper deletion childs
    AddFrame(contentFrame, new TGLayoutHints(kLHintsCenterX| kLHintsExpandX | kLHintsExpandY,2,2,0,2));

    // add embedded canvas to the main window
    fECanvas = new TRootEmbeddedCanvas(0,contentFrame, m_canvasW, m_canvasH);
    contentFrame->AddFrame(fECanvas, new TGLayoutHints(kLHintsLeft | kLHintsExpandX | kLHintsExpandY));

    fCanvas = fECanvas->GetCanvas();
    fCanvas->SetDoubleBuffer(1);
    fCanvas->SetGrid();
    fCanvas->SetLeftMargin(0.15);
    fCanvas->SetBottomMargin(0.15);

    fCanvas->cd();
    pad = new TPad("pad","Pad",0.02,0.02,0.98,0.98);
    pad->Draw();
  
};

void HistoViewer::TryToClose()
{
    // The user try to close the main window,
    //  while a message dialog box is still open.
    printf("Can't close the window '%s' : a message box is still open\n",
           GetWindowName());
};

// Close main window upon click on "X"
void HistoViewer::CloseMainWindow()
{
    //((TServiceClient *)tServiceClient)->Unsubscribe((string)DEMO_TOPIC);
    ((TServiceClient *)tServiceClient)->Cleanup();

    //delete this;
};



void HistoViewer::Subscribe()
{
   
	TStopwatch *stopwatch=new TStopwatch();;
    int count=0;
    void **args=new void *[10];
    args[0]=fCanvas;
    args[1]=pad;
    args[2]=stopwatch;

    /*Callback function to call when a message is received.*/
    ((TCallback *)tCallback)->registerCallback(args, "HistoViewer",
            "callback");
    ((TServiceClient *)tServiceClient)->Subscribe(HISTO_VIEW_TOPIC,tCallback);
    cout<<"Subscribed to response topic"<<endl;
};

//########################### CALL BACK FUNCTION ############################################
/**
 * Callback function. This will handle the displaying of the received images on the canvas.
 * */
void HistoViewer::callback(void **v, char *bytes, int length)
{
    //cout<<"COMING TO CALLBACK FUNCTION"<<endl;
    
    TCanvas *canvas=(TCanvas *)v[0];  
    TPad *pad=(TPad *)v[1];  
    TStopwatch *stpw=(TStopwatch *)v[2];
    
    stpw->Start();
    TImage *img = TImage::Create();
    
    //Read the timestamp comming with the message///////////////////////////
    long long offset=0;
   	long long time=TUtils::ReadLong(bytes,offset);
   	
   	long long current_time=TUtils::GetCurrentTimeMicroseconds();
   	//cout<<"Time in transit = "<<current_time-time<<endl;  
   	//cout<<"Payload length ="<<length<<endl;
  
     	
   	char *b=(bytes+8);
   	  
   
    //img->SetImageBuffer(&bytes, TImage::kXpm);
    //img->SetImageBuffer(&(bytes), TImage::kPng);
    img->SetImageBuffer(&(b), TImage::kPng);
       
    canvas->cd();
    pad->cd();
    img->Draw();
    canvas->Update();
    gSystem->ProcessEvents();
    stpw->Stop();
   
    
    //stpw->Print("m");
    cout<<stpw->RealTime()<<endl;
    delete img;
    //cout<<"Time to render the image = " <<stpw->RealTime()<<endl;
    
    
    

};

void HistoViewer()
{
    gDebug = kFALSE;
    gSystem->Load("clarens.sl");
    /**************************************************************************/
    gSystem->Load("libnbclarens.so"); //Added for nbclarens pubsub capabilities
    /**************************************************************************/
    gROOT->SetStyle("Plain");

    cout<<"Starting Application"<<endl;

    HistoViewer *hw=new HistoViewer(gClient->GetRoot(),200,200);
    hw->Subscribe();

}