// Author: Julian Bunn, Caltech
// Calls local Root analysis, then remote analysis via Clarens

#define Analyse_Eg05_cxx
#include "Analyse_Eg05.h"
#include <TLegend.h>
#include <TPad.h>
#include <TStyle.h>
#include <TCanvas.h>
#include <string.h>

#include <map>

#include <stdio.h>
#include <time.h>


#define MY_PI 3.141592654
#define MAX_WAIT_COUNT 100

#define ANALYSIS_C_FILE 	"Analyse_Eg05.C"
#define ANALYSIS_H_FILE 	"Analyse_Eg05.h"

#define ROOT_DATA 		"born_1.root"
#define OUTPUT_DATA		"histos.root"

#define CITHEP97_CLARENS 	"https://gf3.ucs.indiana.edu:8443/clarens/"
#define ALTAY_CLARENS 	"https://156.56.104.15:8443/clarens/"

#define NB_BROKER_HOST "156.56.104.15"
#define NB_BROKER_PORT 5045

#define NB_ENTITY_ID 6582

#define NB_TEMPLATE_ID 8899
#define DEMO_TOPIC "/clarens/nb/demo/topic"

#define CLOCKS_PER_SEC 1000;


Int_t FALSE = 0;
Int_t TRUE = 1;
using namespace std;

typedef void (* CALLBACK_FUNC)(void **,char*,int);

void callback(void **v,char *bytes,int length);

static int count=0;

void wait ( int seconds )
{
  clock_t endwait;
  endwait = clock () + seconds * CLOCKS_PER_SEC ;
  while (clock() < endwait) {}
};



int Main(){

  TStopwatch stopwatch;
  
  const char *topdir;
  char upload_file[512];
  char output_file[512];
  char tline[10000];
  char trun[4];
  
  cerr << "Load Clarens plugin" << endl;
  gDebug = kFALSE;  
  gSystem->Load("clarens.sl");
  /**************************************************************************/
  gSystem->Load("libnbclarens.so"); //Added for nbclarens pubsub capabilities
  /**************************************************************************/

  Clarens myclarens=Clarens();
  myclarens.Connect(CITHEP97_CLARENS);
  TCSystemDirectory s("/rootdata", CITHEP97_CLARENS, "/rootdata", &myclarens);
  if(!s) {
  	cerr << "Error loading remote system directory TCSystemDirectory" << endl;
	exit(0);
  }
  
  
  Clarens altayClarens=Clarens();
  altayClarens.Connect(ALTAY_CLARENS);
  TCSystemDirectory altaySysDir("/rootdata", ALTAY_CLARENS, "/rootdata", &altayClarens);
  if(!altaySysDir) {
   	cerr << "Error loading remote system directory TCSystemDirectory" << endl;
 	exit(0);
  }
  

  TCanvas *canvas = new TCanvas("canvas","Analysis of Remote Data using Rootlets",0,0,1200,920);
  //TPad *pad1 = new TPad("pad1","Pad1",0.02,0.02,0.48,0.98);
  TPad *pad2 = new TPad("pad2","Pad2",0.02,0.02,0.98,0.98);

  //pad1->Draw();
  pad2->Draw();
    
 // TLegend *localLegend = new TLegend(0.4,0.6,0.9,0.8);
  TLegend *remoteLegend = new TLegend(0.4,0.6,0.9,0.8); 
  

  
  cerr << "Analysis using local .root file" << endl;
  
  strcpy(tline,".x ");
  strcat(tline,ANALYSIS_C_FILE);
  strcat(tline,"(\"");
  strcat(tline,ROOT_DATA);
  strcat(tline,"\",\"");
  strcat(tline,OUTPUT_DATA);
  strcat(tline,"\")");
  
  stopwatch.Start(1);

  cerr << "Executing: " << tline << endl;
  
  gROOT->ProcessLine(tline);

  stopwatch.Stop();
  
  cerr << "Local Analysis: Elapsed " << stopwatch.RealTime() << "secs" << endl;
 
  TFile tflocal(OUTPUT_DATA);
  
  // Find the hRecMass1 histo in the file, and plot it in Pad1  
  //pad1->cd();
  
  TH1F *hRecMass = (TH1F *) tflocal.Get("hRecMass1");
  hRecMass->SetMarkerStyle(21);  
  //  hRecMass->SetFillColor(16);
  // Copy the histogram
  
  TH1F *hRecMassRemote = (TH1F *) hRecMass->Clone();
  hRecMassRemote->SetMarkerStyle(21);
  //  hRecMassRemote->SetFillColor(42);
  
  //hRecMass->Draw();
   
  //canvas->Modified();
  //canvas->Update();
  
  // Load the fit procedure
  
  gROOT->ProcessLine(".L fitMass.C");
  
  // Fit the histogram, overlaying the result
  
  //fitMass(hRecMass,localLegend);
  //localLegend->SetHeader("LOCAL Analysis");
  //localLegend->Draw();
  
  // Plot the clone histogram in Pad2. This histogram will accumulate the
  // results from the remote Rootlet processes


   pad2->cd();

  hRecMassRemote->SetMarkerStyle(21);
//  hRecMassRemote->SetFillColor(42);  
  hRecMassRemote->Draw();
  
  fitMass(hRecMassRemote,remoteLegend);
  
  remoteLegend->SetHeader("REMOTE Analysis");
  remoteLegend->Draw();
   
  canvas->Modified();
  canvas->Update();
   
  s.SetFilter("*.root");
  gDebug=kTRUE;
  s.UpdateContents(kTRUE);

  /**************************************************************************/
  /*It is better to pass the parameters to the Callback function using a TMap.
   * However since Clarens is not an TObject the roocint complains when trying
   * to generate the dictionary. So have to use an array of void pointers
   */
   
  
  //map<string,Clarens*> servers;  
  
  void *args[10];
  args[0]=canvas;
  args[1]=myclarens;
  args[2]=altayClarens;
  args[3]=hRecMassRemote;
  args[4]=remoteLegend;
  
  /*Callback function to call when a message is received.*/  
  //void (* f)(void **v,char *bytes,int length);
  CALLBACK_FUNC f=&callback;
  
  TServiceClient tc;
  tc.Init(NB_BROKER_HOST,NB_BROKER_PORT,NB_ENTITY_ID,NB_TEMPLATE_ID);

  TCallback tbk;
  tbk.registerCallback(f,args);
  tc.Subscribe(DEMO_TOPIC,&tbk);   
    
  gSystem->ProcessEvents();   
  while(true){
		canvas->Update();
		tc.Wait();	
  }
  /*To avoid the UI exiting after fitting the histograms.
   * There should be a better way to do it, but I could not find it :(*/
   
 
   
 // int a;
 // cin>>a;
 //for(;;){
//	gSystem->Sleep(1);
 //   gSystem->ProcessEvents();
 //}
  
  
  tc.Unsubscribe(DEMO_TOPIC);
  tc.Cleanup();	
  /**************************************************************************/
     
  return 0;
}


/**
 * Callback function. This will handle the merging of the 
 * histogram results from the remote processing.
 * */
void callback(void **v,char *bytes,int length){
	
	string msg(bytes,length); 
	cerr<<"Message "<<msg<<endl;

     TCanvas *canvas=(TCanvas *)v[0];
     Clarens *myClarens=(Clarens *)v[1];
     Clarens *altayClarens=(Clarens *)v[2];
     TH1F *hRecMassRemote=(TH1F *)v[3];
     TLegend *remoteLegend=(TLegend *)v[4];
     
     string server(bytes,0,3);
     string shell_histo_name(bytes,4,length-4);
     cerr<<" Server Name = "<<server<<" Histogram File : "<<shell_histo_name<<endl;
     
     Clarens *clarens;
     if(server.compare("gf3")==0){
    	 clarens= myClarens;
     }else if(server.compare("alt")==0){
    	 clarens=altayClarens;
     }else{    	 
    	 cerr<<"Not supported yet"<<endl;
    	 exit(0);
     }        
     
	
     int connect = clarens->Connected();
     cerr << "Connect returned " << connect << endl;
     if(!connect) {
	 	cerr << "Error connecting to Rootlet" << endl;
	    exit(0);
     }
	
	
     TH1F *hr = NULL;
                    
     TCWebFile tfremote((const char *)(shell_histo_name.c_str()), clarens);
     if(!tfremote.IsZombie() &&  tfremote.GetNkeys() > 0){
	     //active_files.AddLast(new TObjString(info->name.Data()));
	     //j++;
	      
	      TH1F *hrm = (TH1F *) tfremote.Get("hRecMass1");
	      hr = (TH1F *) hrm->Clone();
	      hr->Print();
	      cerr << "adding" << endl;
	      hRecMassRemote->Add(hr);
	      hRecMassRemote->SetMarkerStyle(21);
	      //hRecMassRemote->SetFillColor(42);
	
	      //if (j%4==0){
	      cerr << " fitting "<< endl;
	      fitMass(hRecMassRemote,remoteLegend);
	      //}
		  //hRecMassRemote->Draw();              
	      //hRecMass->Draw();
	              
	      cerr << " drawing "<< endl;              
	      remoteLegend->SetHeader("REMOTE Analysis");
	      remoteLegend->Draw();
	      cerr << " updating "<< endl;
	      canvas->Update();
       }   
     
     count++;
     cerr<<"======================================================================="<<endl;
     cerr<<"Received "<<count<<" th update from "<<shell_histo_name<<endl;
     cerr<<"======================================================================="<<endl;
     
};

