//#####################################################################
//## AUTHOR:      Jaliya Ekanakayek (jekanaya@cs.indiana.edu)
//## FILE:        RootUI.C
//##
//## DESCRIPTION: User interface to add Clarens servers and receive
//## 			  Notications when the files are been analysed.
//##
//## THANKS     : Source code published by Tayfun Ince (Tayfun.Ince@cern.ch)
//#####################################################################
#include <stdlib.h>
#include <string.h>
#include <map.h>
#include <vector.h>
#include <iostream>
#include <stdio.h>
#include <time.h>

#include <TROOT.h>
#include <TApplication.h>
#include <TGListBox.h>
#include <TGListTree.h>
#include <TGFSContainer.h>
#include <TGClient.h>
#include <TGFrame.h>
#include <TGIcon.h>
#include <TGLabel.h>
#include <TGButton.h>
#include <TGTextEntry.h>
#include <TGMsgBox.h>
#include <TGTextEdit.h>
#include <TGProgressBar.h>
#include <RQ_OBJECT.h>
#include <TH1.h>
#include <TH2.h>
#include <TRandom.h>
#include <TSystem.h>
#include <TSystemDirectory.h>
#include <TEnv.h>
#include <TFile.h>
#include <TGButton.h>
#include <TGDockableFrame.h>
#include <TRootEmbeddedCanvas.h>
#include <TGStatusBar.h>
#include <TGComboBox.h>
#include <TSystem.h>
#include <TH1D.h>
#include <TH1F.h>
#include <TString.h>
#include <TLegend.h>
#include <TPad.h>
#include <TCanvas.h>
#include <TImage.h>

#define Analyse_Eg05_cxx
#include "Analyse_Eg05.h"

Int_t FALSE = 0;
Int_t TRUE = 1;

#define MY_PI 3.141592654

//Hard-coded configurations===========
#define NB_BROKER_HOST "156.56.104.95";
#define NB_BROKER_PORT 5045;
#define NB_ENTITY_ID_DIV_FACTOR 8000;
#define NB_TEMPLATE_ID 8899;
#define DEMO_TOPIC "/test/clarens/hep";
#define RESPONSE_TOPIC "/test/clarens/hep";
#define SESSION_MGT_TOPIC "/test/session_mangement"

#define ANALYSIS_C_FILE 	"Analyse_Eg05.C";
#define ANALYSIS_H_FILE 	"Analyse_Eg05.h";

#define ROOT_DATA 		"born_1.root";
#define OUTPUT_DATA "output.histo";
#define TMP_HISTO "tmp.histo";

#define GF18_CLARENS	"https://gf18.ucs.indiana.edu:8443/clarens/";
#define GF17_CLARENS 	"https://gf17.ucs.indiana.edu:8443/clarens/";
#define GF15_CLARENS 	"https://gf15.ucs.indiana.edu:8443/clarens/";
#define ROOT_FILES_DIR "/home/clarens/rootfiles/";
#define IMAGE_LOCATION "RootletArchitecture.jpg";
//End=================================

#define BEFORE_MERGE 1;
#define AFTER_MERGE 2;

#define SES_NAME_REQ "ses_name_req";
#define SES_NAME_RES "ses_name_res";
#define SES_DATA_REQ "ses_data_req";
#define SES_DATA_RES "ses_data_res"
#define SES_INFO_SUB "ses_info_sub";
#define MT_JOB_REQ "MT_JOB_REQ";
#define MT_DIR_REQ "MT_DIR_REQ";

using namespace std;

// Class header
class NewRootUI : public TGMainFrame
{

private:

    TGHorizontalFrame *contentFrame;
    TGHorizontalFrame *statusFrame;
    TGVertical3DLine *separator;

    TGVerticalFrame *controlFrame;
    TGHorizontalFrame *ctrBtnFrame;

    //To hold the available clarens servers.
    TGListBox *fListBox;
    TList *fSelected;

    TGHorizontalFrame *frmNumFiles;
    TGTextEntry *numFiles;
    TGLabel *lblNumFiles;
    TGHorizontalFrame *frmSessionName;
    TGTextEntry *sessionName;
    TGLabel *lblSessionName;
    TGTextButton *fLoad;
    TGTextButton *fCreateSession;
    TGTextButton *fAnalyse;

    //The canvas to add the labels for root files.
    TRootEmbeddedCanvas *rootEFileCanvas;
    TCanvas *rootFileCanvas;

    //All the servers initialized
    map<string, void*> clarensServers;
    //Servers of which the files are loaded for processing.
    int currentServerPosInCanvas;
    int num_files_per_rootlet;

    TRootEmbeddedCanvas *fECanvas;
    TGStatusBar *fStatusBar;

    TCanvas *fCanvas;
    TPad *pad;
    
    TImage *img;

    int m_canvasW, m_canvasH;

    void * tServiceClient;
    void * tCallback;
    void *rFileCanvasMgr;
   

public:
    NewRootUI(const TGWindow *p, UInt_t w, UInt_t h);
    virtual ~NewRootUI();
    void CloseMainWindow();

    void Analyse();
    bool LoadFiles(void *clarensServer);
    void Connect();
    void CreateSession();

    void Subscribe();

    void SetupContentFrame();
    void SetupControlFrame();
    void SetupOtherControls();
    int getIndex(vector<string> vec, string val);

    void TryToClose();

    void AddClarensServer(string host);

    void ShowMessageBox(string type, string message);
    string getFileNameFromRemoteURL(string str);

    void callback(void **v, char *bytes, int length);    
    void Tokenize(const string& str,
            vector<string>& tokens,
            const string& delimiters = " ");

};

//#####################################################################
//#####################################################################
// Class constructor
NewRootUI::NewRootUI(const TGWindow *p, UInt_t w, UInt_t h)
{
    // a few gui notes:
    // do not share TGLayoutHints;
    // if using TGCompositeFrame (or other that inherits) as a container,
    // use TGCompositeFrame::SetCleanup(kDeepCleanup) in order to ensure
    // proper deletion of child gui object inside when destructer called;
    // when TGDockableFrame is used, need to delete all child objects
    // individually in the destructor.

    // initialize variables
    currentServerPosInCanvas=0;
    num_files_per_rootlet=0;
    m_canvasW = 720;
    m_canvasH = 540;

    Connect("CloseWindow()", "NewRootUI", this, "CloseMainWindow()");

    //ShowMenu();
    SetupContentFrame();
    SetupControlFrame();
    rFileCanvasMgr=new TFileCanvasManager(rootFileCanvas);

    SetupOtherControls();

    // set main window name and position
    SetWindowName("NBR-Clarens-ROOT");
    MapSubwindows();
    Resize(GetDefaultSize());
   
    // initialize layout of the main window and map the GUI
    Layout();
    MapWindow();

    //Random entity Id
    srand((unsigned) time(0));
    int entityID = (rand() % NB_ENTITY_ID_DIV_FACTOR) + NB_ENTITY_ID_DIV_FACTOR;
    
    //For the publish/subscribe messaging.
    tServiceClient=(TServiceClient *)(new TServiceClient());
    ((TServiceClient *)tServiceClient)->Init(NB_BROKER_HOST,NB_BROKER_PORT,entityID,NB_TEMPLATE_ID);
    tCallback=new TCallback();     
};

void NewRootUI::SetupOtherControls()
{
    // add status bar for messages to the main window
    fStatusBar = new TGStatusBar(this, 50, 10, kHorizontalFrame);
    AddFrame(fStatusBar, new TGLayoutHints(kLHintsExpandX, 2, 2, 0, 2));
    fStatusBar-> SetHeight(30);
    fStatusBar->SetText("NBR-Clarens-ROOT UI.");

};

void NewRootUI::SetupControlFrame()
{
    fListBox = new TGListBox(controlFrame, 100);
    fListBox->Selected(1);
    fListBox->Resize(300, 100);
    fListBox-> SetMultipleSelections(true);
    controlFrame->AddFrame(fListBox, new TGLayoutHints( kLHintsLeft |kLHintsExpandX, 5, 5, 5, 5));

	//Number of Files Text Box
    frmNumFiles = new TGHorizontalFrame(controlFrame, 200, 200);
    frmNumFiles->SetCleanup(kDeepCleanup); // for proper deletion childs
    controlFrame->AddFrame(frmNumFiles, new TGLayoutHints(kLHintsRight,4,4,2,2));

    lblNumFiles=new TGLabel(frmNumFiles, "Number of files per rootlet");
    frmNumFiles->AddFrame(lblNumFiles, new TGLayoutHints( kLHintsLeft , 5, 5, 5, 5));
    
    numFiles=new TGTextEntry(frmNumFiles, "");
    numFiles->SetToolTipText("Number of files per rootlet", 500);
    frmNumFiles->AddFrame(numFiles, new TGLayoutHints( kLHintsLeft, 5, 5, 5, 5));

	//Session Name Text Box
    frmSessionName = new TGHorizontalFrame(controlFrame, 200, 200);
    frmSessionName->SetCleanup(kDeepCleanup); // for proper deletion childs
    controlFrame->AddFrame(frmSessionName, new TGLayoutHints(kLHintsRight,4,4,2,2));

    lblSessionName=new TGLabel(frmSessionName, "Session Name");
    frmSessionName->AddFrame(lblSessionName, new TGLayoutHints( kLHintsLeft, 5, 5, 5, 5));
    
    sessionName=new TGTextEntry(frmSessionName, "");
    sessionName->SetToolTipText("Name for this Analysis Session", 500);
    frmSessionName->AddFrame(sessionName, new TGLayoutHints( kLHintsLeft, 5, 5, 5, 5));


    ctrBtnFrame= new TGHorizontalFrame(controlFrame, 200, 200);
    ctrBtnFrame->SetCleanup(kDeepCleanup);
    controlFrame->AddFrame(ctrBtnFrame, new TGLayoutHints(kLHintsCenterX,4,4,2,2));

    //Setup the controll frame
    fLoad = new TGTextButton(ctrBtnFrame," Connect ");
    fLoad->Connect("Clicked()", "NewRootUI", this, "Connect()");
    ctrBtnFrame->AddFrame(fLoad, new TGLayoutHints(kLHintsLeft,4,4,2,2));

    fCreateSession = new TGTextButton(ctrBtnFrame," Create Session ");
    fCreateSession->Connect("Clicked()", "NewRootUI", this, "CreateSession()");
    ctrBtnFrame->AddFrame(fCreateSession, new TGLayoutHints(kLHintsLeft,4,4,2,2));


    fAnalyse = new TGTextButton(ctrBtnFrame," Analyze ");
    fAnalyse->Connect("Clicked()", "NewRootUI", this, "Analyse()");
    ctrBtnFrame->AddFrame(fAnalyse, new TGLayoutHints(kLHintsLeft,4,4,2,2));

    // add embedded canvas to the main window
    rootEFileCanvas = new TRootEmbeddedCanvas(0,controlFrame, 200,m_canvasH);
    rootEFileCanvas->SetAutoFit(kTRUE);
    controlFrame->AddFrame(rootEFileCanvas, new TGLayoutHints(kLHintsLeft| kLHintsExpandX | kLHintsExpandY));

    //Setup the control frame canvas.
    rootFileCanvas=rootEFileCanvas->GetCanvas();
    rootFileCanvas->SetGrid();
    rootFileCanvas->SetLeftMargin(0.15);
    rootFileCanvas->SetBottomMargin(0.15);
};

void NewRootUI::SetupContentFrame()
{
    //Add a horizontal frame to hold the histogram and the locations sub frames
    contentFrame = new TGHorizontalFrame(this, 200, 200);
    contentFrame->SetCleanup(kDeepCleanup); // for proper deletion childs
    AddFrame(contentFrame, new TGLayoutHints(kLHintsCenterX| kLHintsExpandX | kLHintsExpandY,2,2,0,2));

    // add embedded canvas to the main window
    fECanvas = new TRootEmbeddedCanvas(0,contentFrame, m_canvasW, m_canvasH);
    contentFrame->AddFrame(fECanvas, new TGLayoutHints(kLHintsLeft | kLHintsExpandX | kLHintsExpandY));

    fCanvas = fECanvas->GetCanvas();
    fCanvas->SetDoubleBuffer(1);
    fCanvas->SetGrid();
    fCanvas->SetLeftMargin(0.15);
    fCanvas->SetBottomMargin(0.15);

    fCanvas->cd();
	img = TImage::Open(IMAGE_LOCATION);
    pad = new TPad("pad","Pad",0.02,0.02,0.98,0.98);
    pad->Draw();
    pad->cd();   
    img->Draw();      

    separator = new TGVertical3DLine(contentFrame);
    contentFrame->AddFrame(separator, new TGLayoutHints(kLHintsLeft | kLHintsExpandY,2,2,0,0));

    controlFrame=new TGVerticalFrame(contentFrame, 200, 200);
    controlFrame->SetCleanup(kDeepCleanup); // for proper deletion childs
    contentFrame->AddFrame(controlFrame, new TGLayoutHints(kLHintsLeft | kLHintsExpandY,2,2,0,2));
}
;

void NewRootUI::TryToClose()
{
    // The user try to close the main window,
    //  while a message dialog box is still open.
    printf("Can't close the window '%s' : a message box is still open\n",  GetWindowName());
};

/*
 * Utility method to display a message box.
 * 
 * */
void NewRootUI::ShowMessageBox(string type, string message)
{

    int i, buttons, retval;
    EMsgBoxIcon icontype = kMBIconStop;
    buttons = kMBDismiss;

    // Since the message dialog box is created, we disable the
    // signal/slot communication mechanism, in order to ensure we
    // can't close the this window while the message box is open.
    Disconnect("CloseMainWindow()");
    Connect("CloseWindow()", "NewRootUI", this, "TryToClose()");
    //TODO: How to delete this pointer?
    new TGMsgBox(gClient->GetRoot(), this,type.c_str(), message.c_str(), icontype, buttons, &retval);
    Disconnect("CloseWindow()");
    Connect("CloseWindow()", "NewRootUI", this, "CloseMainWindow()");

};

/*
 * Add a clarens object to available servers after connecting to it.
 * These clarens objects are already connected to a particular server
 * can directly use after verifying the connectivity.
 * */
void NewRootUI::AddClarensServer(string host)
{
    TClarensServer *server=new TClarensServer(host);
    clarensServers.insert(make_pair(host, (void *)server ));
    fListBox->NewEntry(host.c_str());
};

/**
 * Subscribe to the topic to receive histogram files.
 * 
 */
void NewRootUI::Subscribe()
{
	   char tline[10000];

	   TLegend *legend = new TLegend(0.4,0.6,0.9,0.8);
	   legend->SetHeader("REMOTE Analysis");

	   strcpy(tline, ".x ");
	   strcat(tline, ANALYSIS_C_FILE);
	   strcat(tline, "(\"");
	   strcat(tline, ROOT_DATA);
	   strcat(tline, "\",\"");
	   strcat(tline, OUTPUT_DATA);
	   strcat(tline, "\")");

	   gROOT->ProcessLine(tline);

	   TFile *tflocal=new TFile(OUTPUT_DATA);
	   TH1F *hRecMass = (TH1F *) tflocal->Get("hRecMass1");
	   hRecMass->SetMarkerStyle(21);

	   gROOT->ProcessLine(".L fitMass.C");
	   fitMass(hRecMass, legend);	

	   //Send the pointers to the Callback method. ROOT callback mechanism is
	   //reflection based,so the callback method cannot access the class members
	   //therefore we need to pass it the pointer to the class to access its memebers.
	   void **args=new void *[3];
	   args[0]=this;
	   args[1]=legend;
	   args[2]=hRecMass;

    /*Callback function to call when a message is received.*/
    ((TCallback *)tCallback)->registerCallback(args, "NewRootUI","callback");
    ((TServiceClient *)tServiceClient)->Subscribe(DEMO_TOPIC,tCallback);
};

/**
 * Compose the analysis tasks and subimt them to the available Clarens servers.
 * 
 */
void NewRootUI::Analyse()
{
    fStatusBar->SetText("Starts Submitting Analysis Jobs...");
    gSystem->ProcessEvents();
   
    map<string,void*>::iterator iter;
    for (iter = clarensServers.begin(); iter != clarensServers.end(); iter++)
    {
        TClarensServer *server=(TClarensServer *)(iter->second);
        if (server->isLoaded())
        {
            int num_files=server->getNumRootFile();
            int mod=num_files%num_files_per_rootlet;
            int div=num_files/num_files_per_rootlet;
            int iterations=div;
            if (mod!=0)
            {
                iterations=div+1;
            }    
                
            for (int n=0; n<iterations; n++)
            {
                int startIndex = n*num_files_per_rootlet;
                int endIndex = (n+1)*num_files_per_rootlet-1;
                if (endIndex>=num_files)
                {
                    endIndex=num_files-1;
                }
                
                //name of the histogram
                stringstream histo_name;
                histo_name<<"Rootlet_";
                histo_name<<n;
                histo_name<<".histo";

                TRootletHelper helper;
                helper.setScriptName(ANALYSIS_C_FILE);
                helper.setOutputFileName(histo_name.str());
                helper.setResponseTopic(RESPONSE_TOPIC);

                for (int i=startIndex; i<=endIndex; i++)
                {
                    helper.addInputFileName(server->getRootFile(i));                 
                }

                //TODO CINT does not allow dynamic array sizes.
                const int estimate_tline_size=40000;

                char tline[estimate_tline_size];
                tline[0]='\0';
                if (helper.getTLineString(tline))
                {
                   	stringstream buff;  
                   	buff<< MT_JOB_REQ;            	
                	buff<<"|";
                	buff<<tline;
                	buff<<"|";
                	buff<<server->getUrl();
                        	
                	((TServiceClient *)tServiceClient)->Publish(server->getUrl(),(char*)(buff.str().c_str()),strlen(buff.str().c_str()));
   					buff.clear();//Clear the buffer.

                }
                else
                {
                    cout<<"Error:Generating the Executable Description"<<endl;
                    exit(0);
                }
            }
        }	        
        cout << "INFO: Finished submitting requests to: " << iter->first << endl;
    }    
}
;


/*
 * Initializes the User Interface.
 * 
 * */
void NewRootUI::Connect()
{
    fStatusBar-> ChangeBackground(46);
    fStatusBar->SetText("Connecting ...");
    gSystem->ProcessEvents();

    Subscribe();
    bool filesLoaded=true;

    if (clarensServers.size()==0)
    {
        ShowMessageBox("Information", "No servers to connect");
        return;
    }
    else
    {
        map<string,void *>::iterator iter;
        for (iter = clarensServers.begin(); iter != clarensServers.end(); iter++)
        {
            filesLoaded=LoadFiles(iter->second);
        }
    }

    ((TFileCanvasManager *)(this->rFileCanvasMgr))->draw();

    fStatusBar->SetText("Sucessfully Connected.");
    gSystem->ProcessEvents();

};

/**
 * Initlaize a session for this data analysis.
 */
void NewRootUI::CreateSession(){
	
	int num=0;
    std::istringstream ss(numFiles->GetText());
    ss >> num;
    if (num==0)
    {
        ShowMessageBox("Information", "Please specify number of files per rootlet.");
        return;
    }   	
	this->num_files_per_rootlet=num;
	
	string session_name(sessionName->GetText());
    if (session_name.length()==0)
    {
        ShowMessageBox("Information", "Please specify a name for this session.");
        return;
    }   	
	
	TFileCanvasManager *mgr=(TFileCanvasManager *)(this->rFileCanvasMgr);
	mgr->setNumFilePerRootlet(num);	
	mgr->updateCanvas();
	
	 stringstream buff;
	 buff<<SES_INFO_SUB;
	 buff<<"|";
	 buff<<session_name;
	 buff<<"|";
	 buff<<fListBox->GetNumberOfEntries();
	 buff<<"|";
	 buff<<num_files_per_rootlet;	 
	 map<string,void *>::iterator iter;
     for (iter = clarensServers.begin(); iter != clarensServers.end(); iter++)
     {
        TClarensServer *server=(TClarensServer*)(iter->second);
        buff<<"|";
        buff<<server->getUrl();
        buff<<"#";
        buff<<server->getNumRootFile();
     }
	
	//cout<<"SES_INFO_SUB "<<buff.str()<<endl;
    ((TServiceClient *)tServiceClient)->Publish(SESSION_MGT_TOPIC,(char*)(buff.str().c_str()),strlen(buff.str().c_str()));    
    
};

/*
 * Load root files in  a clarens server for processing
 * to a vector named connectedServers. Connect to the server
 * and retrieve the list of root file names available for 
 * processing in it and populate the rootFileCanvas.
 * */
bool NewRootUI::LoadFiles(void *server)
{
    TClarensServer *clarensServer=(TClarensServer *)server;
    string host=clarensServer->getUrl();

    //TODO have to check if the files are already added or not. Also to faciliate removal
    //files are already loaded or not.
    if (clarensServer->isLoaded())
    {
        ShowMessageBox("Info", "Files are already loaded.");
        return false;
    }
    //Now add the Clarens object to loadedServers.
    clarensServer->setLoaded(true);
    
    TUtils tu;
       
    srand(tu.GetCurrentTimeMicroseconds());
    int eId = (rand() % NB_ENTITY_ID_DIV_FACTOR) + NB_ENTITY_ID_DIV_FACTOR; 
    
    TCall call;
    call.Init(NB_BROKER_HOST,NB_BROKER_PORT,eId,787878);
    string resTopic="/mt_dir_req/response/"+host; 
    string data=MT_DIR_REQ+resTopic;//first 10 bytes are for the message type
    
    char *res;
    int  resLen;
	call.SendReceive(host,resTopic,data.c_str(),data.length(),&res,&resLen);
		
	string response(res,resLen);
	vector<string> files;
	Tokenize(response, files,",");
	call.Cleanup();       
   
    int num_files=files.size();
   // cerr << "There are " << num_files<< " files" << endl;

    if (num_files==0)
    {
        ShowMessageBox("Info", "No root files available for processing at the host ");
        return true;
    }

    int endOfHTTPS=host.find(":", 0)+3;
    int lengthOfFirstSegment=host.find(".", 0) -endOfHTTPS;
    string lblStr=host.substr(endOfHTTPS, lengthOfFirstSegment);
    ((TFileCanvasManager *)(this->rFileCanvasMgr))->addServer(host,lblStr,num_files);

    for (int i=0; i<num_files; i++)
    {
        string fileName=ROOT_FILES_DIR;
        fileName.append(files.at(i));
        clarensServer->addRootFile(fileName);
    }

};

//#####################################################################
// Class destructorl,
NewRootUI::~NewRootUI()
{
    Cleanup();
};

//#####################################################################
// Close main window upon click on "X"
void NewRootUI::CloseMainWindow()
{
    ((TServiceClient *)tServiceClient)->Cleanup();
    delete this;
};

string NewRootUI::getFileNameFromRemoteURL(string str)
{
    int lastIndex=0;
    bool found=false;
    string::size_type loc;
    for (int i=0; i<str.length(); i++)
    {
        loc = str.find("/", i);
        if (loc != string::npos)
        {
            lastIndex=loc;
        }
        else
        {
            found =true;
        }
    }
    if (found)
    {
        return str.substr(lastIndex+1);
    }
    else
    {
        ShowMessageBox("Information", "Could not find the file name.");
        //TODO correct this.
        return "";
    }

};


void NewRootUI::callback(void **v, char *bytes, int length)
{    
    NewRootUI *ui=(NewRootUI *)v[0];
    TFileCanvasManager *mgr=(TFileCanvasManager *)(ui->rFileCanvasMgr);
    TLegend *remoteLegend=(TLegend *)v[1];
    TH1F *hRecMassRemote=(TH1F *)v[2];    
    
    //Read the timestamp comming with the message
    long long offset=0;
	long long time=TUtils::ReadLong(bytes,offset); 
       
    const int header_len=TUtils::ReadInt(bytes+offset,offset);
    const int new_header_len=header_len;
    char header[new_header_len];
    TUtils::CopyBytes(bytes+offset,header,header_len,offset);	
    string header_str(header,header_len);
    
    ////////////////////////////Update the Canvase/////////////////////////////////////////
    string::size_type loc = header_str.find("#", 0);
    if (loc == string::npos)
    {
        cerr<<"Could not understand the message "<<endl;
        return;
    }

    string serverUrl(header_str.c_str(), 0, loc);
    string shell_histo_name(header_str.c_str(), loc+1, length-loc-1);
    
    string fileName=getFileNameFromRemoteURL(shell_histo_name);    
   
    int num_pos_begin=fileName.find_first_of("_",0) +1; //Rootlet_x.histo;
    int num_pos_end=fileName.find_first_of(".",0); //Rootlet_x.histo;
    string str_num=fileName.substr(num_pos_begin,(num_pos_end-num_pos_begin));
    
	istringstream buffer(str_num);
	int rootlet_num;
	buffer >> rootlet_num;        
    mgr->updateRootlet(serverUrl,rootlet_num, BEFORE_MERGE);    
    /////////////////////////////End updating Canvas///////////////////////////////////////    
       
    ui->fCanvas->cd();
    ui->pad->cd();
    
    ofstream myFile (TMP_HISTO, ios::trunc|ios::out | ios::binary);
    myFile.write (bytes+offset, length-offset);
    myFile.flush();
    myFile.close();                
    
    //Check if the file exists else use the option NEW for the constructor.  
    TFile *tfile=new TFile(TMP_HISTO);    
    TH1F *hrm = (TH1F *) tfile->Get("hRecMass1");
   
    hRecMassRemote->Add(hrm);
    hRecMassRemote->SetMarkerStyle(21);
   
    fitMass(hRecMassRemote, remoteLegend);

    remoteLegend->Draw();
    ui->fCanvas->Update();    
 	mgr->updateRootlet(serverUrl,rootlet_num, AFTER_MERGE);    
    delete tfile;
};

//########################### END OF CALL BACK FUNCTION #######################################

/**
 * A utility method to tokanize a string.
 */
void Tokenize(const string& str,
                      vector<string>& tokens,
                      const string& delimiters = " ")
{
    // Skip delimiters at beginning.
    string::size_type lastPos = str.find_first_not_of(delimiters, 0);
    // Find first "non-delimiter".
    string::size_type pos     = str.find_first_of(delimiters, lastPos);

    while (string::npos != pos || string::npos != lastPos)
    {
        // Found a token, add it to the vector.
        tokens.push_back(str.substr(lastPos, pos - lastPos));
        // Skip delimiters.  Note the "not_of"
        lastPos = str.find_first_not_of(delimiters, pos);
        // Find next "non-delimiter"
        pos = str.find_first_of(delimiters, lastPos);
    }
}

//#####################################################################
// Launch GUI
void NewRootUI()
{
    gDebug = kFALSE;
    /**************************************************************************/
    gSystem->Load("libnbROOT.so"); //Added for nbclarens pubsub capabilities
    /**************************************************************************/
    gROOT->SetStyle("Plain");

    NewRootUI *ui=new NewRootUI(gClient->GetRoot(),200,200);
    ui->AddClarensServer(GF17_CLARENS);
    ui->AddClarensServer(GF15_CLARENS);
    ui->AddClarensServer(GF18_CLARENS);

}
//#####################################################################


