//#####################################################################
//## AUTHOR:      Jaliya Ekanakayek (jekanaya@cs.indiana.edu)
//## FILE:        RootUI.C
//##
//## DESCRIPTION: User interface to add Clarens servers and receive
//## 			  Notications when the files are been analysed.
//##
//## THANKS     : Source code published by Tayfun Ince (Tayfun.Ince@cern.ch)
//#####################################################################
#include <stdlib.h>

#include <TROOT.h>
#include <TApplication.h>
#include <TVirtualX.h>
#include <TVirtualPadEditor.h>
#include <TGResourcePool.h>
#include <TGListBox.h>
#include <TGListTree.h>
#include <TGFSContainer.h>
#include <TGClient.h>
#include <TGFrame.h>
#include <TGIcon.h>
#include <TGLabel.h>
#include <TGButton.h>
#include <TGTextEntry.h>
#include <TGNumberEntry.h>
#include <TGMsgBox.h>
#include <TGMenu.h>
#include <TGCanvas.h>
#include <TGComboBox.h>
#include <TGTab.h>
#include <TGSlider.h>
#include <TGDoubleSlider.h>
#include <TGFileDialog.h>
#include <TGTextEdit.h>
#include <TGShutter.h>
#include <TGProgressBar.h>
#include <TGColorSelect.h>
#include <RQ_OBJECT.h>
#include <TCanvas.h>
#include <TColor.h>
#include <TH1.h>
#include <TH2.h>
#include <TRandom.h>
#include <TSystem.h>
#include <TSystemDirectory.h>
#include <TEnv.h>
#include <TFile.h>
#include <TKey.h>
#include <TGDockableFrame.h>
#include <TGFontDialog.h>

#include <TGButton.h>
#include <TGDockableFrame.h>
#include <TRootEmbeddedCanvas.h>
#include <TGMenu.h>
#include <TPostScript.h>
#include <TCanvas.h>
#include <TStyle.h>
#include <TGStatusBar.h>
#include <TGComboBox.h>
#include <TGFileDialog.h>
#include <TGraph.h>
#include <TSystem.h>
#include <TH1D.h>
#include <TH1F.h>
#include <TString.h>
#include <TImage.h>

#define Analyse_Eg05_cxx
#include "Analyse_Eg05.h"
#include <TLegend.h>
#include <TPad.h>
#include <TStyle.h>
#include <TCanvas.h>
#include <string.h>
#include <map.h>
#include <vector.h>
#include <iostream>

#include <stdio.h>
#include <time.h>

Int_t FALSE = 0;
Int_t TRUE = 1;

#define MY_PI 3.141592654

#define NB_BROKER_HOST "156.56.104.95";
#define NB_BROKER_PORT 5045;
#define NB_ENTITY_ID_DIV_FACTOR 8000;
#define NB_TEMPLATE_ID 8899;
//#define DEMO_TOPIC "/clarens/nb/demo/topic"
#define DEMO_TOPIC "/test/clarens/hep";
#define SESSION_MGT_TOPIC "/test/session_mangement"

#define ANALYSIS_C_FILE 	"Analyse_Eg05.C";
#define ANALYSIS_H_FILE 	"Analyse_Eg05.h";

#define MAX_FILES_PER_ROOTLET 10;

#define ROOT_DATA 		"born_1.root";
#define OUTPUT_FILE_PREFIX  "histos.root";
#define OUTPUT_DATA "output.histo";

#define TMP_HISTO "tmp.histo";

#define RESPONSE_TOPIC "/test/clarens/hep";

#define GF3_CLARENS 	"https://gf3.ucs.indiana.edu:8443/clarens/";
#define GF18_CLARENS	"https://gf18.ucs.indiana.edu:8443/clarens/";
#define GF17_CLARENS 	"https://gf17.ucs.indiana.edu:8443/clarens/";
#define GF15_CLARENS 	"https://gf15.ucs.indiana.edu:8443/clarens/";
#define ALTAY_CLARENS 	"https://156.56.104.15:8443/clarens/";
#define ANALYSIS_FILE "Analyse_Eg05_Clarens_indiana.C";
#define ANALYSIS_SCRIPT "root-executer.py";
//#define ANALYSIS_SCRIPT "perl_test.pl"

#define IMAGE_LOCATION "RootletArchitecture.jpg";

#define CLOCKS_PER_SEC 1000;

#define BEFORE_MERGE 1;
#define AFTER_MERGE 2;

#define SES_NAME_REQ "ses_name_req";
#define SES_NAME_RES "ses_name_res";
#define SES_DATA_REQ "ses_data_req";
#define SES_DATA_RES "ses_data_res"
#define SES_INFO_SUB "ses_info_sub";


using namespace std;

// Class header
class RootUI : public TGMainFrame
{

private:

    TGHorizontalFrame *contentFrame;
    TGHorizontalFrame *statusFrame;
    TGVertical3DLine *separator;

    TGVerticalFrame *controlFrame;
    TGHorizontalFrame *ctrBtnFrame;

    //To hold the available clarens servers.
    TGListBox *fListBox;
    TList *fSelected;

    TGHorizontalFrame *frmNumFiles;
    TGTextEntry *numFiles;
    TGLabel *lblNumFiles;
    TGHorizontalFrame *frmSessionName;
    TGTextEntry *sessionName;
    TGLabel *lblSessionName;
    TGTextButton *fLoad;
    TGTextButton *fCreateSession;
    TGTextButton *fAnalyse;

    //The canvas to add the labels for root files.
    TRootEmbeddedCanvas *rootEFileCanvas;
    TCanvas *rootFileCanvas;

    //All the servers initialized
    map<string, void*> clarensServers;
    //Servers of which the files are loaded for processing.
    int currentServerPosInCanvas;
    int num_files_per_rootlet;

    TRootEmbeddedCanvas *fECanvas;
    //TGHorizontalFrame    	*fHFrame;
    TGStatusBar *fStatusBar;

    TCanvas *fCanvas;
    TPad *pad;

    int m_canvasW, m_canvasH;

    void * tServiceClient;
    void * tCallback;
    void *rFileCanvasMgr;

    TImage *img;
    
    //long long startTime;
    //int rootletCount;
    //int numRootlet;
    
    //TStopwatch *stopwatch;
    

public:
    RootUI(const TGWindow *p, UInt_t w, UInt_t h);
    virtual ~RootUI();
    void CloseMainWindow();

    void Analyse();
    bool LoadFiles(void *clarensServer);
    void Connect();
    void CreateSession();

    void Subscribe();

    void SetupContentFrame();
    void SetupControlFrame();
    void SetupOtherControls();
    int getIndex(vector<string> vec, string val);

    void TryToClose();

    void AddClarensServer(string host);

    void ShowMessageBox(string type, string message);
    string getFileNameFromRemoteURL(string str);

    void callback(void **v, char *bytes, int length);
    
    void sendImage(string topic,char* data, int len);
    
    void Tokenize(const string& str,
            vector<string>& tokens,
            const string& delimiters = " ");

};

//#####################################################################
//#####################################################################
// Class constructor
RootUI::RootUI(const TGWindow *p, UInt_t w, UInt_t h)
{
    // a few gui notes:
    // do not share TGLayoutHints;
    // if using TGCompositeFrame (or other that inherits) as a container,
    // use TGCompositeFrame::SetCleanup(kDeepCleanup) in order to ensure
    // proper deletion of child gui object inside when destructer called;
    // when TGDockableFrame is used, need to delete all child objects
    // individually in the destructor.

    // initialize variables
    currentServerPosInCanvas=0;
    num_files_per_rootlet=0;
    //m_canvasW = 720;
    //m_canvasH = 540;
    
    m_canvasW = 420;
    m_canvasH = 360;
    
    
    //startTime=0;
    //rootletCount=0;
    //numRootlet=0;

    Connect("CloseWindow()", "RootUI", this, "CloseMainWindow()");

    //ShowMenu();
    SetupContentFrame();
    SetupControlFrame();
    rFileCanvasMgr=new TFileCanvasManager(rootFileCanvas);

    SetupOtherControls();

    // set main window name and position
    SetWindowName("NBR-Clarens-ROOT");
    MapSubwindows();
    Resize(GetDefaultSize());
    //MoveResize(0,0,GetMaxHeight(),GetMaxWidth());

    // initialize layout of the main window and map the GUI
    Layout();
    MapWindow();

    //Random entity Id
    srand((unsigned) time(0));
    int entityID = (rand() % NB_ENTITY_ID_DIV_FACTOR) + NB_ENTITY_ID_DIV_FACTOR;
    
    //For the publish/subscribe messaging.
    tServiceClient=(TServiceClient *)(new TServiceClient());
    ((TServiceClient *)tServiceClient)->Init(NB_BROKER_HOST,NB_BROKER_PORT,entityID,NB_TEMPLATE_ID);
    tCallback=new TCallback();
    
    //stopwatch=new TStopwatch();    
}
;

void RootUI::SetupControlFrame()
{
    fListBox = new TGListBox(controlFrame, 100);
    fListBox->Selected(1);
    fListBox->Resize(300, 100);
    fListBox-> SetMultipleSelections(true);
    controlFrame->AddFrame(fListBox, new TGLayoutHints( kLHintsLeft |kLHintsExpandX, 5, 5, 5, 5));

	//Number of Files Text Box
    frmNumFiles = new TGHorizontalFrame(controlFrame, 200, 200);
    frmNumFiles->SetCleanup(kDeepCleanup); // for proper deletion childs
    controlFrame->AddFrame(frmNumFiles, new TGLayoutHints(kLHintsRight,4,4,2,2));

    lblNumFiles=new TGLabel(frmNumFiles, "Number of files per rootlet");
    frmNumFiles->AddFrame(lblNumFiles, new TGLayoutHints( kLHintsLeft , 5, 5, 5, 5));
    
    numFiles=new TGTextEntry(frmNumFiles, "");
    numFiles->SetToolTipText("Number of files per rootlet", 500);
    frmNumFiles->AddFrame(numFiles, new TGLayoutHints( kLHintsLeft, 5, 5, 5, 5));

	//Session Name Text Box
    frmSessionName = new TGHorizontalFrame(controlFrame, 200, 200);
    frmSessionName->SetCleanup(kDeepCleanup); // for proper deletion childs
    controlFrame->AddFrame(frmSessionName, new TGLayoutHints(kLHintsRight,4,4,2,2));

    lblSessionName=new TGLabel(frmSessionName, "Session Name");
    frmSessionName->AddFrame(lblSessionName, new TGLayoutHints( kLHintsLeft, 5, 5, 5, 5));
    
    sessionName=new TGTextEntry(frmSessionName, "");
    sessionName->SetToolTipText("Name for this Analysis Session", 500);
    frmSessionName->AddFrame(sessionName, new TGLayoutHints( kLHintsLeft, 5, 5, 5, 5));


    ctrBtnFrame= new TGHorizontalFrame(controlFrame, 200, 200);
    ctrBtnFrame->SetCleanup(kDeepCleanup);
    controlFrame->AddFrame(ctrBtnFrame, new TGLayoutHints(kLHintsCenterX,4,4,2,2));

    //Setup the controll frame
    fLoad = new TGTextButton(ctrBtnFrame," Connect ");
    fLoad->Connect("Clicked()", "RootUI", this, "Connect()");
    ctrBtnFrame->AddFrame(fLoad, new TGLayoutHints(kLHintsLeft,4,4,2,2));

    fCreateSession = new TGTextButton(ctrBtnFrame," Create Session ");
    fCreateSession->Connect("Clicked()", "RootUI", this, "CreateSession()");
    ctrBtnFrame->AddFrame(fCreateSession, new TGLayoutHints(kLHintsLeft,4,4,2,2));


    fAnalyse = new TGTextButton(ctrBtnFrame," Analyze ");
    fAnalyse->Connect("Clicked()", "RootUI", this, "Analyse()");
    ctrBtnFrame->AddFrame(fAnalyse, new TGLayoutHints(kLHintsLeft,4,4,2,2));

    // add embedded canvas to the main window
    //rootEFileCanvas = new TRootEmbeddedCanvas(0,controlFrame, m_canvasW, m_canvasH);
    rootEFileCanvas = new TRootEmbeddedCanvas(0,controlFrame, 200,m_canvasH);
    rootEFileCanvas->SetAutoFit(kTRUE);
    controlFrame->AddFrame(rootEFileCanvas, new TGLayoutHints(kLHintsLeft| kLHintsExpandX | kLHintsExpandY));

    //Setup the control frame canvas.
    rootFileCanvas=rootEFileCanvas->GetCanvas();
    rootFileCanvas->SetGrid();
    rootFileCanvas->SetLeftMargin(0.15);
    rootFileCanvas->SetBottomMargin(0.15);


}
;

void RootUI::SetupContentFrame()
{
    //Add a horizontal frame to hold the histogram and the locations sub frames
    contentFrame = new TGHorizontalFrame(this, 200, 200);
    contentFrame->SetCleanup(kDeepCleanup); // for proper deletion childs
    AddFrame(contentFrame, new TGLayoutHints(kLHintsCenterX| kLHintsExpandX | kLHintsExpandY,2,2,0,2));

    // add embedded canvas to the main window
    fECanvas = new TRootEmbeddedCanvas(0,contentFrame, m_canvasW, m_canvasH);
    contentFrame->AddFrame(fECanvas, new TGLayoutHints(kLHintsLeft | kLHintsExpandX | kLHintsExpandY));

    fCanvas = fECanvas->GetCanvas();
    fCanvas->SetDoubleBuffer(1);
    fCanvas->SetGrid();
    fCanvas->SetLeftMargin(0.15);
    fCanvas->SetBottomMargin(0.15);

    fCanvas->cd();
	img = TImage::Open(IMAGE_LOCATION);
	//float d = 0.40;
    pad = new TPad("pad","Pad",0.02,0.02,0.98,0.98);
    //pad = new TPad("pad","Pad",0.05, 0.55, 0.05+d*img->GetWidth()/img->GetHeight(), 0.95);
    pad->Draw();
    pad->cd();   
    img->Draw();
      

    separator = new TGVertical3DLine(contentFrame);
    contentFrame->AddFrame(separator, new TGLayoutHints(kLHintsLeft | kLHintsExpandY,2,2,0,0));

    controlFrame=new TGVerticalFrame(contentFrame, 200, 200);
    controlFrame->SetCleanup(kDeepCleanup); // for proper deletion childs

    contentFrame->AddFrame(controlFrame, new TGLayoutHints(kLHintsLeft | kLHintsExpandY,2,2,0,2));
}
;

void RootUI::TryToClose()
{
    // The user try to close the main window,
    //  while a message dialog box is still open.
    printf("Can't close the window '%s' : a message box is still open\n",
           GetWindowName());
}
;

/*
 * Utility method to display a message box.
 * */
void RootUI::ShowMessageBox(string type, string message)
{
    //cout<<"Showing message"<<endl;

    int i, buttons, retval;
    EMsgBoxIcon icontype = kMBIconStop;
    buttons = kMBDismiss;

    // Since the message dialog box is created, we disable the
    // signal/slot communication mechanism, in order to ensure we
    // can't close the this window while the message box is open.
    Disconnect("CloseMainWindow()");
    Connect("CloseWindow()", "RootUI", this, "TryToClose()");
    //TODO: How to delete this pointer?
    new TGMsgBox(gClient->GetRoot(), this,type.c_str(), message.c_str(), icontype, buttons, &retval);
    Disconnect("CloseWindow()");
    Connect("CloseWindow()", "RootUI", this, "CloseMainWindow()");

}
;

/*
 * Add a clarens object to available servers after connecting to it.
 * These clarens objects are already connected to a particular server
 * can directly use after verifying the connectivity.
 * */
void RootUI::AddClarensServer(string host)
{
    TClarensServer *server=new TClarensServer(host);
    clarensServers.insert(make_pair(host, (void *)server ));
    fListBox->NewEntry(host.c_str());
}
;

void RootUI::SetupOtherControls()
{
    // add status bar for messages to the main window
    fStatusBar = new TGStatusBar(this, 50, 10, kHorizontalFrame);
    AddFrame(fStatusBar, new TGLayoutHints(kLHintsExpandX, 2, 2, 0, 2));
    fStatusBar-> SetHeight(30);
    fStatusBar->SetText("NBR-Clarens-ROOT UI.");

}
;

void RootUI::Subscribe()
{
	   char tline[10000];

	   TLegend *legend = new TLegend(0.4,0.6,0.9,0.8);
	   legend->SetHeader("REMOTE Analysis");
	  // cerr << "Analysis using local .root file" << endl;

	   strcpy(tline, ".x ");
	   strcat(tline, ANALYSIS_C_FILE);
	   strcat(tline, "(\"");
	   strcat(tline, ROOT_DATA);
	   strcat(tline, "\",\"");
	   strcat(tline, OUTPUT_DATA);
	   strcat(tline, "\")");

	   //cerr << "Executing Local Analysis: " << tline << endl;
	   gROOT->ProcessLine(tline);

	   TFile *tflocal=new TFile(OUTPUT_DATA);
	   TH1F *hRecMass = (TH1F *) tflocal->Get("hRecMass1");
	   hRecMass->SetMarkerStyle(21);

	   gROOT->ProcessLine(".L fitMass.C");
	   fitMass(hRecMass, legend);	
	

//	gROOT->ProcessLine(".L fitMass.C");
//
//    TLegend *legend = new TLegend(0.4,0.6,0.9,0.8);
//    legend->SetHeader("REMOTE Analysis");
//        
//    TH1F *hRecMass = new TH1F("hRecMass1","REMOTE Analysis",100,0,4.4);

    
   // vector<long long> *time_delay=new vector<long long>(); 
	   
	    void **args=new void *[3];
	     args[0]=this;
	      args[1]=legend;
	      args[2]=hRecMass;
	   
    
//    int count=0;
//    void **args=new void *[15];
//    args[0]=&clarensServers;
//    args[1]=fCanvas;
//    args[2]=hRecMass;
//    args[3]=legend;
//    args[4]=rootFileCanvas;
//    args[5]=pad;
//    args[6]=rFileCanvasMgr;  
//    args[7]=tServiceClient; 
    //args[8]=stopwatch;
    //args[8]=time_delay;
    //args[9]=&rootletCount;
    //args[10]=&numRootlet;

    /*Callback function to call when a message is received.*/
    ((TCallback *)tCallback)->registerCallback(args, "RootUI",
            "callback");
    ((TServiceClient *)tServiceClient)->Subscribe(DEMO_TOPIC,tCallback);
    //cout<<"Subscribed to response topic"<<endl;
};



void RootUI::Analyse()
{
    fStatusBar->SetText("Starts Submitting Analysis Jobs...");
    gSystem->ProcessEvents();
    
   // this->stopwatch->Start();
   
    map<string,void*>::iterator iter;
    for (iter = clarensServers.begin(); iter != clarensServers.end(); iter++)
    {
        TClarensServer *server=(TClarensServer *)(iter->second);

        if (server->isLoaded())
        {
            int num_files=server->getNumRootFile();
            int mod=num_files%num_files_per_rootlet;
            int div=num_files/num_files_per_rootlet;
            int iterations=div;
            if (mod!=0)
            {
                iterations=div+1;
            }
            
            //this->numRootlet=iterations*clarensServers.size();
            
            for (int n=0; n<iterations; n++)
            {
                int startIndex = n*num_files_per_rootlet;
                int endIndex = (n+1)*num_files_per_rootlet-1;
                if (endIndex>=num_files)
                {
                    endIndex=num_files-1;
                }
                
               // cout<<"Num Files "<<num_files<<" Start Index "<<startIndex<<" End Index "<<endIndex<<endl;
       

                //histogram name
                stringstream histo_name;
                histo_name<<"Rootlet_";
                histo_name<<n;
                histo_name<<".histo";

                TRootletHelper helper;
                helper.setScriptName(ANALYSIS_C_FILE);
                helper.setOutputFileName(histo_name.str());
                helper.setResponseTopic(RESPONSE_TOPIC);

                for (int i=startIndex; i<=endIndex; i++)
                {
                    helper.addInputFileName(server->getRootFile(i));
                    //cout<<"Adding files "<<server->getRootFile(i)<<endl;
                }

                //TODO CINT does not allow dynamic array sizes.
                //const int estimate_tline_size=server->getRootFile(startIndex).size()*num_files;
                const int estimate_tline_size=40000;

                char tline[estimate_tline_size];
                tline[0]='\0';
                if (helper.getTLineString(tline))
                {
                   	stringstream buff;
//                	buff<<"root -b -l -q '";
//                	buff<<ANALYSIS_FILE;
//                	buff<<"(\"";
//                	buff<<tline;
//                	buff<<"\",\"";
//                	buff<<server->getUrl();
//                	buff<<"\")' &";
//                	cout<<"The buffer is "<<buff.str()<<endl<<endl;
//                	gSystem->Exec(buff.str().c_str());
                	
                	buff<<"MT_JOB_REQ|";
                	buff<<tline;
                	buff<<"|";
                	buff<<server->getUrl();
                	
                	//cout<<"The buffer is "<<buff.str()<<endl<<endl;
                	
                	((TServiceClient *)tServiceClient)->Publish(server->getUrl(),(char*)(buff.str().c_str()),strlen(buff.str().c_str()));
                	
                    buff.clear();//Clear the buffer.

                }
                else
                {
                    cout<<"Error:Generating the Executable Description"<<endl;
                    exit(0);
                }
            }
        }		
        
        cout << "INFO: Finished submitting requests to: " << iter->first << endl;
    }
   
    
}
;

/*
 * Returns the index of the element if it is already in the vector.
 * Else return -1.
 * */
int RootUI::getIndex(vector<string> vec, string val)
{
    int size=vec.size();
    for (int i=0; i<size; i++)
    {
        if((vec.at(i)).compare(val.c_str())==0)
            )
        {
            return i;
        }
    }
    return -1;
}
;

/*
 * Simplify the UI functions
 * 
 * */
void RootUI::Connect()
{
    fStatusBar-> ChangeBackground(46);
    fStatusBar->SetText("Connecting ...");
    gSystem->ProcessEvents();

    Subscribe();
    bool filesLoaded=true;

    if (clarensServers.size()==0)
    {
        ShowMessageBox("Information", "No servers to connect");
        return;
    }
    else
    {
        map<string,void *>::iterator iter;
        for (iter = clarensServers.begin(); iter != clarensServers.end(); iter++)
        {
            filesLoaded=LoadFiles(iter->second);
        }
    }

    ((TFileCanvasManager *)(this->rFileCanvasMgr))->draw();

    fStatusBar->SetText("Sucessfully Connected.");
    gSystem->ProcessEvents();

}
;

void RootUI::CreateSession(){
	
	int num=0;
    std::istringstream ss(numFiles->GetText());
    ss >> num;
    if (num==0)
    {
        ShowMessageBox("Information",
                       "Please specify number of files per rootlet.");
        return;
    }   	
	this->num_files_per_rootlet=num;
	
	string session_name(sessionName->GetText());
    if (session_name.length()==0)
    {
        ShowMessageBox("Information",
                       "Please specify a name for this session.");
        return;
    }   	
	//cout<<"Session Name = "<<session_name<<endl;
	
	TFileCanvasManager *mgr=(TFileCanvasManager *)(this->rFileCanvasMgr);
	mgr->setNumFilePerRootlet(num);	
	mgr->updateCanvas();
	
	//Now send the session info submission message
	
	 stringstream buff;
	 buff<<SES_INFO_SUB;
	 buff<<"|";
	 buff<<session_name;
	 buff<<"|";
	 buff<<fListBox->GetNumberOfEntries();
	 buff<<"|";
	 buff<<num_files_per_rootlet;	 
	 map<string,void *>::iterator iter;
     for (iter = clarensServers.begin(); iter != clarensServers.end(); iter++)
     {
        TClarensServer *server=(TClarensServer*)(iter->second);
        buff<<"|";
        buff<<server->getUrl();
        buff<<"#";
        buff<<server->getNumRootFile();
     }
	
	//cout<<"SES_INFO_SUB "<<buff.str()<<endl;
    ((TServiceClient *)tServiceClient)->Publish(SESSION_MGT_TOPIC,(char*)(buff.str().c_str()),strlen(buff.str().c_str()));    
    
};

/*
 * Load root files in  a clarens server for processing
 * to a vector named connectedServers. Connect to the server
 * and retrieve the list of root file names available for 
 * processing in it and populate the rootFileCanvas.
 * */
bool RootUI::LoadFiles(void *server)
{
    TClarensServer *clarensServer=(TClarensServer *)server;
    string host=clarensServer->getUrl();

    //TODO have to check if the files are already added or not. Also to faciliate removal
    //files are already loaded or not.

    if (clarensServer->isLoaded())
    {
        ShowMessageBox("Info", "Files are already loaded.");
        return false;
    }
    //Now add the Clarens object to loadedServers.
    clarensServer->setLoaded(true);
    
    //============================================================================
    TUtils tu;
       
    srand(tu.GetCurrentTimeMicroseconds());
    int eId = (rand() % NB_ENTITY_ID_DIV_FACTOR) + NB_ENTITY_ID_DIV_FACTOR; 
    
    TCall call;
    call.Init(NB_BROKER_HOST,NB_BROKER_PORT,eId,787878);
    string resTopic="/mt_dir_req/response/"+host; 
    string data="MT_DIR_REQ"+resTopic;//first 10 bytes are for the message type
    
    char *res;
    int  resLen;
	call.SendReceive(host,resTopic,data.c_str(),data.length(),&res,&resLen);
		
	string response(res,resLen);
	vector<string> files;
	Tokenize(response, files,",");
	call.Cleanup();
    
    //============================================================================
   
   
    int num_files=files.size();
   // cerr << "There are " << num_files<< " files" << endl;

    if (num_files==0)
    {
        ShowMessageBox("Info",
                       "No root files available for processing at the host ");
        return true;
    }

    int endOfHTTPS=host.find(":", 0)+3;
    int lengthOfFirstSegment=host.find(".", 0) -endOfHTTPS;
    string lblStr=host.substr(endOfHTTPS, lengthOfFirstSegment);
    ((TFileCanvasManager *)(this->rFileCanvasMgr))->addServer(host,lblStr,num_files);

    for (int i=0; i<num_files; i++)
    {
        string fileName="/home/clarens/rootfiles/";
        fileName.append(files.at(i));
        clarensServer->addRootFile(fileName);
    }

}
;

//#####################################################################
// Class destructorl,
RootUI::~RootUI()
{
    // delete TGDockableFrame childs
    //delete fMenuBar;
    //delete fMenuPS;
    //delete fMenuRootUI;

    Cleanup();
}
;

//#####################################################################
// Close main window upon click on "X"
void RootUI::CloseMainWindow()
{
    //((TServiceClient *)tServiceClient)->Unsubscribe((string)DEMO_TOPIC);
    ((TServiceClient *)tServiceClient)->Cleanup();

    delete this;
}
;

string RootUI::getFileNameFromRemoteURL(string str)
{
    int lastIndex=0;
    bool found=false;
    string::size_type loc;
    for (int i=0; i<str.length(); i++)
    {
        loc = str.find("/", i);
        if (loc != string::npos)
        {
            //cout << "Found Omega at " << loc << endl;
            lastIndex=loc;
        }
        else
        {
            //cout << "Didn't find Omega" << endl;
            found =true;
        }
    }
    if (found)
    {
        return str.substr(lastIndex+1);
    }
    else
    {
        ShowMessageBox("Information", "Could not find the file name.");
        //TODO correct this.
        return "";
    }

}
;

//########################### CALL BACK FUNCTION ############################################
/**
 * Callback function. This will handle the merging of the 
 * histogram results from the remote processing.
 * */
 /*
void RootUI::callback(void **v, char *bytes, int length)
{
    cout<<"COMING TO CALLBACK FUNCTION"<<endl;

    map<string, void*> *servers=(map<string, void*> *)v[0];
    TCanvas *canvas=(TCanvas *)v[1];
    TH1F *hRecMassRemote=(TH1F *)v[2];
    TLegend *remoteLegend=(TLegend *)v[3];
    TCanvas *fileCanvas=(TCanvas *)v[4];
    TPad *padx=(TPad *)v[5];
    TFileCanvasManager *mgr=(TFileCanvasManager *)v[6];
    TServiceClient *tc=(TServiceClient *)v[7];   

    string msg(bytes, length);
    string::size_type loc = msg.find("#", 0);
    if (loc == string::npos)
    {
        cerr<<"Could not understand the message "<<endl;
        return;
    }

    string serverUrl(bytes, 0, loc);
    string shell_histo_name(bytes, loc+1, length-loc-1);
    cerr<<" Server Name = "<<serverUrl<<" Histogram File : "<<shell_histo_name   <<endl;

    TClarensServer *server;
    Clarens *clarens;
    map<string,void *>::iterator iter = servers->find(serverUrl);
    if (iter != servers->end() )
    {
        server=(TClarensServer *)(iter->second);
        clarens=(Clarens*)(server->getServer());
    }
    else
    {
        ShowMessageBox("Information", "Could not find the server.");
        return;
    }
   
    int connect = clarens->Connected();
    cerr << "Connect returned " << connect << endl;
    if (!connect)
    {
        cerr << "Error connecting to Rootlet" << endl;
        exit(0);
    }

    TH1F *hr = NULL;
    TCWebFile tfremote((const char *)(shell_histo_name.c_str()), clarens);

    string fileName=getFileNameFromRemoteURL(shell_histo_name);    
   
    int num_pos_begin=fileName.find_first_of("_",0) +1; //Rootlet_x.histo;
    int num_pos_end=fileName.find_first_of(".",0); //Rootlet_x.histo;
    string str_num=fileName.substr(num_pos_begin,(num_pos_end-num_pos_begin));
    
	istringstream buffer(str_num);
	int rootlet_num;
	buffer >> rootlet_num;        
    mgr->updateRootlet(serverUrl,rootlet_num, BEFORE_MERGE);
    

    if (!tfremote.IsZombie() && tfremote.GetNkeys() > 0)      
    {
        //active_files.AddLast(new TObjString(info->name.Data()));
        canvas->cd();
        padx->cd();

        TH1F *hrm = (TH1F *) tfremote.Get("hRecMass1");
        cerr << "adding " << endl;
        hRecMassRemote->Add(hrm);
        hRecMassRemote->SetMarkerStyle(21);

        cerr << " fitting "<< endl;
        fitMass(hRecMassRemote, remoteLegend);


        cerr << " drawing "<< endl;
        remoteLegend->Draw();
        cerr << " updating "<< endl;
        canvas->Update();

    }
    else
    {
        cout<<"NOT PROCESSSING "<<endl;
    }
    mgr->updateRootlet(serverUrl,rootlet_num, AFTER_MERGE);
    
    cout<<"start sendign histogram as an image"<<endl;
    TImage *img = TImage::Create();
    img->FromPad(padx);
   
    
    char *buf;
    int sz;
    img->GetImageBuffer(&buf, &sz, TImage::kXpm); //raw buffer       
    //img->WriteImage(img_name.c_str(),TImage::kPng);      
    
    cout<<"Size of the image ="<<sz<<endl;
    string topic="/test/shared/display";
  	tc->Publish(topic,buf,sz);
    
    delete[] buf;    
    delete img;
    cout<<"End saving as an image"<<endl;
   
    cerr
    <<"======================================================================="
    <<endl;
    cerr<<"Received update from "<<shell_histo_name<<endl;
    cerr
    <<"======================================================================="
    <<endl;

}; */

void RootUI::callback(void **v, char *bytes, int length)
{
    //cout<<"COMING TO CALLBACK FUNCTION"<<endl;
    map<string, void*> *servers=(map<string, void*> *)v[0];
    TCanvas *canvas=(TCanvas *)v[1];
    TH1F *hRecMassRemote=(TH1F *)v[2];
    TLegend *remoteLegend=(TLegend *)v[3];
    TCanvas *fileCanvas=(TCanvas *)v[4];
    TPad *padx=(TPad *)v[5];
    TFileCanvasManager *mgr=(TFileCanvasManager *)v[6];
    TServiceClient *tc=(TServiceClient *)v[7];   
    //TStopwatch *stpw=(TStopwatch *)v[8];
    //vector<long long> *times=(vector<long long> *)v[8];
    //int * count=(int *)v[9];
    //int * numR=(int *)v[10];

    
    RootUI *ui=(RootUI *)v[0];
    TFileCanvasManager *mgr=(TFileCanvasManager *)(ui->rFileCanvasMgr);
    TLegend *remoteLegend=(TLegend *)v[1];
    TH1F *hRecMassRemote=(TH1F *)v[2];
    
    
    //Read the timestamp comming with the message///////////////////////////
    long long offset=0;
	long long time=TUtils::ReadLong(bytes,offset); 

       
    const int header_len=TUtils::ReadInt(bytes+offset,offset);
    const int new_header_len=header_len;
    //cout<<"Header length ="<<new_header_len<<endl;
    char header[new_header_len];
    TUtils::CopyBytes(bytes+offset,header,header_len,offset);	
    string header_str(header,header_len);

    
    ////////////////////////////Update the Canvase/////////////////////////////////////////
    string::size_type loc = header_str.find("#", 0);
    if (loc == string::npos)
    {
        cerr<<"Could not understand the message "<<endl;
        return;
    }

    string serverUrl(header_str.c_str(), 0, loc);
    string shell_histo_name(header_str.c_str(), loc+1, length-loc-1);
    
    string fileName=getFileNameFromRemoteURL(shell_histo_name);    
   
    int num_pos_begin=fileName.find_first_of("_",0) +1; //Rootlet_x.histo;
    int num_pos_end=fileName.find_first_of(".",0); //Rootlet_x.histo;
    string str_num=fileName.substr(num_pos_begin,(num_pos_end-num_pos_begin));
    
	istringstream buffer(str_num);
	int rootlet_num;
	buffer >> rootlet_num;        
    mgr->updateRootlet(serverUrl,rootlet_num, BEFORE_MERGE);
    
    /////////////////////////////End updating Canvas///////////////////////////////////////    
   
    
    ui->fCanvas->cd();
    ui->pad->cd();
    
    ofstream myFile (TMP_HISTO, ios::trunc|ios::out | ios::binary);
    myFile.write (bytes+offset, length-offset);
    myFile.flush();
    myFile.close();        
         
    
    //Check if the file exists else use the option NEW for the constructor.  
    TFile *tfile=new TFile(TMP_HISTO);    
    //tfile->WriteBuffer(bytes+offset,length-offset);   

    TH1F *hrm = (TH1F *) tfile->Get("hRecMass1");
   
    hRecMassRemote->Add(hrm);
    hRecMassRemote->SetMarkerStyle(21);
   
    fitMass(hRecMassRemote, remoteLegend);
    //fitMass(hrm, remoteLegend);

    remoteLegend->Draw();
    ui->fCanvas->Update();    
    
    
    
    mgr->updateRootlet(serverUrl,rootlet_num, AFTER_MERGE);
    
    delete tfile;
    
 
    
    /*
    
    if((*count)==(*numR)){
    	stpw->Stop();
    	cout<<"!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"<<endl;
    	cout<<stpw->RealTime()<<endl;
    	cout<<"!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"<<endl;
    	
    }
    
    */
    
    /*
    
    stpw->Start(); 
    //Publishing the screen as an image
    TImage *img = TImage::Create();
    img->FromPad(padx);   
    
    char *buf=new char[50000]; //CHECK THIS CODE ??????????????????????????????????????/
      
    int sz;
    img->GetImageBuffer(&buf, &sz, TImage::kPng); //raw buffer       
    //img->WriteImage(img_name.c_str(),TImage::kPng);  
    
    
    char *nbuf=new char[50000];
    long long offset=0;
    TUtils::WriteLong(nbuf,time,offset); //Write the time for benchmarks.   
    for(int i=0;i<sz;i++){
    	nbuf[offset+i]=buf[i];
    }
    
           
    //cout<<"Size of the image ="<<sz<<endl;
    string topic="/test/shared/display";
  	tc->Publish(topic,nbuf,sz+offset);
  	stpw->Stop();
  	cout<<stpw->RealTime()<<endl;
    
    delete[] buf;  
    delete[] nbuf;   
    delete img;
   
		*/
   
   
	   
    
    /*
    
    
    string::size_type loc = msg.find("#", 0);
    if (loc == string::npos)
    {
        cerr<<"Could not understand the message "<<endl;
        return;
    }

    string serverUrl(bytes, 0, loc);
    string shell_histo_name(bytes, loc+1, length-loc-1);
    cerr<<" Server Name = "<<serverUrl<<" Histogram File : "<<shell_histo_name   <<endl;

    TClarensServer *server;
    Clarens *clarens;
    map<string,void *>::iterator iter = servers->find(serverUrl);
    if (iter != servers->end() )
    {
        server=(TClarensServer *)(iter->second);
        clarens=(Clarens*)(server->getServer());
    }
    else
    {
        ShowMessageBox("Information", "Could not find the server.");
        return;
    }
   
    int connect = clarens->Connected();
    cerr << "Connect returned " << connect << endl;
    if (!connect)
    {
        cerr << "Error connecting to Rootlet" << endl;
        exit(0);
    }

    TH1F *hr = NULL;
    TCWebFile tfremote((const char *)(shell_histo_name.c_str()), clarens);

    string fileName=getFileNameFromRemoteURL(shell_histo_name);    
   
    int num_pos_begin=fileName.find_first_of("_",0) +1; //Rootlet_x.histo;
    int num_pos_end=fileName.find_first_of(".",0); //Rootlet_x.histo;
    string str_num=fileName.substr(num_pos_begin,(num_pos_end-num_pos_begin));
    
	istringstream buffer(str_num);
	int rootlet_num;
	buffer >> rootlet_num;        
    mgr->updateRootlet(serverUrl,rootlet_num, BEFORE_MERGE);
    

    if (!tfremote.IsZombie() && tfremote.GetNkeys() > 0)      
    {
        //active_files.AddLast(new TObjString(info->name.Data()));
        canvas->cd();
        padx->cd();

        TH1F *hrm = (TH1F *) tfremote.Get("hRecMass1");
        cerr << "adding " << endl;
        hRecMassRemote->Add(hrm);
        hRecMassRemote->SetMarkerStyle(21);

        cerr << " fitting "<< endl;
        fitMass(hRecMassRemote, remoteLegend);


        cerr << " drawing "<< endl;
        remoteLegend->Draw();
        cerr << " updating "<< endl;
        canvas->Update();

    }
    else
    {
        cout<<"NOT PROCESSSING "<<endl;
    }
    mgr->updateRootlet(serverUrl,rootlet_num, AFTER_MERGE);
    
    cout<<"start sendign histogram as an image"<<endl;
    TImage *img = TImage::Create();
    img->FromPad(padx);
   
    
    char *buf;
    int sz;
    img->GetImageBuffer(&buf, &sz, TImage::kXpm); //raw buffer       
    //img->WriteImage(img_name.c_str(),TImage::kPng);      
    
    cout<<"Size of the image ="<<sz<<endl;
    string topic="/test/shared/display";
  	tc->Publish(topic,buf,sz);
    
    delete[] buf;    
    delete img;
    cout<<"End saving as an image"<<endl;
   
    cerr
    <<"======================================================================="
    <<endl;
    cerr<<"Received update from "<<shell_histo_name<<endl;
    cerr
    <<"======================================================================="
    <<endl;
    
    */

};


void 
RootUI::sendImage(string topic,char* data,int len){
	 ((TServiceClient *)tServiceClient)->Publish(topic,data,len);	

};

//########################### END OF CALL BACK FUNCTION #######################################



void Tokenize(const string& str,
                      vector<string>& tokens,
                      const string& delimiters = " ")
{
    // Skip delimiters at beginning.
    string::size_type lastPos = str.find_first_not_of(delimiters, 0);
    // Find first "non-delimiter".
    string::size_type pos     = str.find_first_of(delimiters, lastPos);

    while (string::npos != pos || string::npos != lastPos)
    {
        // Found a token, add it to the vector.
        tokens.push_back(str.substr(lastPos, pos - lastPos));
        // Skip delimiters.  Note the "not_of"
        lastPos = str.find_first_not_of(delimiters, pos);
        // Find next "non-delimiter"
        pos = str.find_first_of(delimiters, lastPos);
    }
}


//#####################################################################
//#####################################################################
// Launch GUI
void RootUI()
{
    gDebug = kFALSE;
    //gSystem->Load("clarens.sl");
    /**************************************************************************/
    gSystem->Load("libnbROOT.so"); //Added for nbclarens pubsub capabilities
    /**************************************************************************/

    gROOT->SetStyle("Plain");
    
    
   
    

    //cout<<"start sleeping"<<endl;
    //TUtils::Sleep(10000);
    //cout<<"End Sleeping"<<endl;
	
    RootUI *ui=new RootUI(gClient->GetRoot(),200,200);
    //ui->AddClarensServer(ALTAY_CLARENS);
    ui->AddClarensServer(GF17_CLARENS);
    ui->AddClarensServer(GF15_CLARENS);
    ui->AddClarensServer(GF18_CLARENS);
    
    
    //Image testings ==================================
    /*

     TServiceClient *tsc=new TServiceClient();
  	 tsc->Init(NB_BROKER_HOST,NB_BROKER_PORT,8697,NB_TEMPLATE_ID);


	long long offset=0;
	long long time=TUtils::GetCurrentTimeMicroseconds();
   

    TImage *img = TImage::Create();

    string img_name="Rootlet_0.histo.png";
    img->ReadImage(img_name.c_str(),TImage::kPng);
        
    char *buf=new char[50000];
    char *nbuf=new char[50000];
    int sz=0;
   // TUtils::WriteLong(buf,time,offset);    
   // cout<<"offset is "<<offset<<endl;
   
    //img->GetImageBuffer(c, &sz, TImage::kXpm); //raw buffer
    img->GetImageBuffer(&buf, &sz, TImage::kPng); //raw buffer
    
    *nbuf='A';
    *(nbuf+1)='B';
    for(int i=0;i<sz;i++){
    	nbuf[2+i]=buf[i];
    }
    
    
    
    //**c='A';
    //*(buf+offset+1)='B';
    //*(buf+offset+2)='C';
    
     for(int i=0;i<sz+2;i++)
    {
    	printf("%d",*(nbuf+i));
    }
    
    cout<<endl<<"Size of the image ="<<sz<<endl;
    string topic="/test/shared/display";
    tsc->Publish(topic,nbuf,sz+2);	
    
    delete[] buf;   
    
    */
    //=================================================
    
    
    
    
    
    
    

    /*for(int i=0;i<10;i++){
        TRootletHelper helper;
        helper.setScriptName("AnalysisScript.C");
        helper.setOutputFileName("output.histo");
        helper.setResponseTopic("/test/clarens/hep");
        helper.addInputFileName("born1.root");
        helper.addInputFileName("born2.root");
        helper.addInputFileName("born3.root");
        
        char tline[10000];
        tline[0]='\0';
        if(helper.getTLineString(tline)){
        	cout<<tline<<endl;
        }else{
        	cout<<"Error"<<endl;
        }
       }*/




}
//#####################################################################


