//#####################################################################
//## AUTHOR:      Tayfun Ince (Tayfun.Ince@cern.ch)
//## FILE:        XMonitoring.C
//## MODIFIED:    2007/05/16  <-- thanks to Ilka Antcheva for bug fixes
//## DESCRIPTION: Template macro to build a simple GUI.
//#####################################################################

#include <TROOT.h>
#include <RQ_OBJECT.h>
#include <TGClient.h> 
#include <TGFrame.h>
#include <TGCanvas.h>
#include <TGButton.h> 
#include <TGDockableFrame.h> 
#include <TRootEmbeddedCanvas.h> 
#include <TGMenu.h> 
#include <TPostScript.h>
#include <TCanvas.h>
#include <TStyle.h> 
#include <TGStatusBar.h> 
#include <TGComboBox.h> 
#include <TGFileDialog.h> 
#include <TGraph.h> 
#include <TSystem.h> 
#include <TH1D.h>
#include <TString.h>

using namespace std;

//#####################################################################
//#####################################################################
// Class header
class XMon
{

  RQ_OBJECT("XMon") // necessary if not inherite from ROOT class

 protected:
  enum EPopupMenuEntry
    { // enumerators for signal/slot mechanism of main menu entries
      M_PS_OPEN,
      M_PS_CLOSE,
      M_PS_PRINT,
      M_PS_DELETE,

      M_XMON_ENT1,
      M_XMON_ENT2,
    };

 private:
  TGMainFrame          *fMain;
  TGHorizontalFrame    *contentFrame;
  
  TGVertical3DLine 	   *separator; 
  
  
  TGDockableFrame      *fMenuDock;
  TRootEmbeddedCanvas  *fECanvas;
  TGMenuBar            *fMenuBar;
  TGPopupMenu          *fMenuPS, *fMenuXMon;
  TGHorizontalFrame    *fHFrame;
  
  TGVerticalFrame      *fVFrame1;
  TGVerticalFrame      *fVFrame2;
  
  TGComboBox           *fCombo;
  TGTextButton         *fDraw, *fSave, *fPrintThis;
  TGStatusBar          *fStatusBar;

  TCanvas              *fCanvas;
  TPostScript          *m_psFile;
  TString              m_psFileName;

  int                  m_canvasW, m_canvasH;

 public:
  XMon(const TGWindow *p, UInt_t w, UInt_t h);
  virtual ~XMon();
  void CloseMainWindow();

  void Draw();
  void Save();
  void PrintThis();

  void HandleMainMenu(Int_t id);
};

//#####################################################################
//#####################################################################
// Class constructor
XMon::XMon(const TGWindow *p, UInt_t w, UInt_t h)
{
  // a few gui notes:
  // do not share TGLayoutHints;
  // if using TGCompositeFrame (or other that inherits) as a container,
  // use TGCompositeFrame::SetCleanup(kDeepCleanup) in order to ensure
  // proper deletion of child gui object inside when destructer called;
  // when TGDockableFrame is used, need to delete all child objects
  // individually in the destructor.

  // initialize variables
  m_canvasW = 200;
  m_canvasH = 200;

  // start building gui
  // create a top level window
  fMain = new TGMainFrame(p,w,h,kHorizontalFrame);

  // connect(signal, receiver_class, receiver, slot)
  fMain->Connect("CloseWindow()", "XMon", this, "CloseMainWindow()");

  // add dockable frame to the main window
  fMenuDock = new TGDockableFrame(fMain);
  fMain->AddFrame(fMenuDock, new TGLayoutHints(kLHintsExpandX, 2, 2, 2, 2));
  fMenuDock->SetWindowName("X Monitoring Menu");

  // create "Postscript" menu
  fMenuPS = new TGPopupMenu(gClient->GetRoot());
  fMenuPS->AddEntry("Open", M_PS_OPEN);
  fMenuPS->AddEntry("Close", M_PS_CLOSE);
  fMenuPS->AddSeparator();
  fMenuPS->AddEntry("Print", M_PS_PRINT);
  fMenuPS->AddSeparator();
  fMenuPS->AddEntry("Delete", M_PS_DELETE);
  fMenuPS->Connect("Activated(Int_t)", "XMon", this, "HandleMainMenu(Int_t)");

  // create "XMon" menu
  fMenuXMon = new TGPopupMenu(gClient->GetRoot());
  fMenuXMon->AddEntry("Entry 1", M_XMON_ENT1);
  fMenuXMon->AddEntry("Entry 2", M_XMON_ENT2);
  fMenuXMon->Connect("Activated(Int_t)", "XMon", this, "HandleMainMenu(Int_t)");

  // add menu bar with popup menus to the dockable frame
  fMenuBar = new TGMenuBar(fMenuDock, 1, 1, kHorizontalFrame);
  fMenuDock->AddFrame(fMenuBar, new TGLayoutHints(kLHintsTop | kLHintsExpandX));
  fMenuBar->AddPopup("&PostScript", fMenuPS, new TGLayoutHints(kLHintsTop | kLHintsLeft, 0, 0, 2, 2));
  fMenuBar->AddPopup("&XMon", fMenuXMon, new TGLayoutHints(kLHintsTop | kLHintsLeft, 0, 0, 2, 2));
  
  //Add a horizontal frame to hold the histogram and the locations sub frames
  contentFrame = new TGHorizontalFrame(fMain, 200, 200);
  contentFrame->SetCleanup(kDeepCleanup);  // for proper deletion childs
  fMain->AddFrame(contentFrame, new TGLayoutHints(kLHintsCenterX,2,2,0,2));
  
  /*
  // add a vertcal frame for combo box and text buttons to the main window
  fVFrame1 = new TGVerticalFrame(contentFrame, 50, 10);
  fVFrame1->SetCleanup(kDeepCleanup);  // for proper deletion childs
  contentFrame->AddFrame(fVFrame1, new TGLayoutHints(kLHintsLeft |kLHintsExpandX | kLHintsExpandY ,2,2,0,2));
  
  // Separator
  separator = new TGVertical3DLine(contentFrame);
  contentFrame->AddFrame(separator, new TGLayoutHints(kLHintsRight | kLHintsExpandY));
  
  
  // add a vertcal frame for combo box and text buttons to the main window
  fVFrame2 = new TGVerticalFrame(contentFrame, 150, 10);
  fVFrame2->SetCleanup(kDeepCleanup);  // for proper deletion childs
  contentFrame->AddFrame(fVFrame2, new TGLayoutHints(kLHintsLeft,2,2,0,2));
  
  // add embedded canvas to the main window
  fECanvas = new TRootEmbeddedCanvas(0,fVFrame1, m_canvasW, m_canvasH);
  fVFrame1->AddFrame(fECanvas, new TGLayoutHints(kLHintsExpandX | kLHintsExpandY));
  fCanvas = fECanvas->GetCanvas();
  
  // plot a graph into canvas
  fCanvas->cd();
  double x[20], y[20];
  int n = 20;
  for(int i=0; i<n; i++){
    x[i] = i*0.1;
    y[i] = 10*sin(x[i]+0.2);
  }
  TGraph *gr = new TGraph(n,x,y);
  gr->Draw("AC*");
  fCanvas->Update();
  
  */

  //add a horizontal frame for combo box and text buttons to the main window
  fHFrame = new TGHorizontalFrame(fMain, 50, 10);
  fHFrame->SetCleanup(kDeepCleanup);  // for proper deletion childs
  fMain->AddFrame(fHFrame, new TGLayoutHints(kLHintsCenterX,2,2,0,2));
   
  // combo box
  fCombo = new TGComboBox(fHFrame);
  fCombo->Resize(100,20);
  fHFrame->AddFrame(fCombo, new TGLayoutHints(kLHintsExpandY,4,4,2,2));
  // fill combo box
  fCombo->AddEntry("Histogram A", 0);
  fCombo->AddEntry("Histogram B", 1);
  fCombo->Select(0); 
  // text buttons
  fDraw = new TGTextButton(fHFrame," Draw ");
  fDraw->Connect("Clicked()", "XMon", this, "Draw()");
  fHFrame->AddFrame(fDraw, new TGLayoutHints(kLHintsExpandY,4,4,2,2));
  fSave = new TGTextButton(fHFrame," Save ");
  fSave->Connect("Clicked()", "XMon", this, "Save()");
  fHFrame->AddFrame(fSave, new TGLayoutHints(kLHintsExpandY,4,4,2,2));
  fPrintThis = new TGTextButton(fHFrame," PrintThis ");
  fPrintThis->Connect("Clicked()", "XMon", this, "PrintThis()");
  fHFrame->AddFrame(fPrintThis, new TGLayoutHints(kLHintsExpandY,4,4,2,2));

  // add status bar for messages to the main window
  fStatusBar = new TGStatusBar(fMain, 50, 10, kHorizontalFrame);
  fMain->AddFrame(fStatusBar, new TGLayoutHints(kLHintsExpandX, 2, 2, 0, 2));
  fStatusBar->SetText("Welcome to X monitoring.");

  // set main window name and position
  fMain->SetWindowName("X Monitoring"); 
  fMain->MapSubwindows(); 
  fMain->Resize(fMain->GetDefaultSize());
  fMain->MoveResize(228,0);
  fMain->SetWMPosition(228,0);

  // initialize layout of the main window and map the GUI
  fMain->Layout();
  fMain->MapWindow();
}

//#####################################################################
// Class destructor
XMon::~XMon()
{
  // delete TGDockableFrame childs
  delete fMenuBar;
  delete fMenuPS;
  delete fMenuXMon;
  
  fMain->Cleanup();
}

//#####################################################################
// Close main window upon click on "X"
void XMon::CloseMainWindow()
{
  delete this;
}



//#####################################################################
// Draw histo selected from combobox 
void XMon::Draw()
{
  gStyle->SetOptStat("emr"); // show only nentries, mean and rms
  gDirectory->DeleteAll(); // delete prev objects in memory

  // get selected combo box entry
  TGTextLBEntry *tEntry = (TGTextLBEntry*)fCombo->GetSelectedEntry();
  TString histoTitle = tEntry->GetText()->GetString();

  if(histoTitle == "Histogram A"){
    fStatusBar->SetText("Plotting " + histoTitle);
    TH1D *h1d = new TH1D("h1d", histoTitle, 10, -0.5, 9.5);
    for(int i=0; i<10; i++){
      double x = i;
      double weight = i+1;
      h1d->Fill(x, weight);
    }
    fCanvas->cd();
    h1d->Draw();
    fCanvas->Update();
  }
  else if(histoTitle == "Histogram B"){
    fStatusBar->SetText(histoTitle + " DOES NOT EXIST");
  }
}

//#####################################################################
// Save canvas in a ps file and keep it open, so can save more
void XMon::Save()
{
  if(m_psFile == 0){
    TString dir(".");
    TGFileInfo fi;
    const char *filetypes[] = {"PostScript files", "*.ps", 0, 0};
    fi.fFileTypes = filetypes;
    fi.fIniDir    = StrDup(dir);
    new TGFileDialog(gClient->GetRoot(), fMain, kFDSave, &fi);
    dir = fi.fIniDir;
    if(!fi.fFilename){
      fStatusBar->SetText("NO FILE SELECTED OR NAME TYPED!");
      return;
    } else{
      m_psFileName = fi.fFilename;
      m_psFile = new TPostScript(m_psFileName,112);
      m_psFile->Off();
    }
  }
  m_psFile->On();
  m_psFile->NewPage();
  fCanvas->Draw();

  TString psFileBaseName = gSystem->BaseName(m_psFile->GetName());
  fStatusBar->SetText("Canvas is saved in FILE: " + psFileBaseName +
		      " and file is left open for future saves in this session." + 
		      " CLOSE FILE BEFORE EXIT.");
  m_psFile->Off(); 
}

//#####################################################################
// Print current active canvas to the printer directly
void XMon::PrintThis()
{
  TString temp = "temp.ps";
  TPostScript ps(temp,112);
  ps.NewPage();
  fCanvas->Draw();
  ps.Close();

  // CERN Bat40 1d cor
  TString printCommand = "xprint -h -p40-1d-cor" + temp;
  gSystem->Exec(printCommand);
  fStatusBar->SetText("Canvas(es) is(are) sent to PRINTER: 40-1d-cor");
  gSystem->Exec("rm -f " + temp);
}

//#####################################################################
// Execute lines between "case ... break" 
// upon click on given top menu entry
void XMon::HandleMainMenu(Int_t id)
{
  const char *filetypes[] = {"PostScript files", "*.ps", 0, 0};
  TGFileInfo fi;
  
  switch (id) {
  case M_PS_OPEN:
    {
    if(m_psFile != 0){
      fStatusBar->SetText("FILE: " + m_psFileName + " IS OPEN!");
      break;
    }
    TString dir(".");
    fi.fFileTypes = filetypes;
    fi.fIniDir    = StrDup(dir);
    new TGFileDialog(gClient->GetRoot(), fMain, kFDSave, &fi);
    dir = fi.fIniDir;
    if(!fi.fFilename){
      fStatusBar->SetText("NO FILE SELECTED OR NAME TYPED!");
      break;
    } else{
      m_psFileName = fi.fFilename;
      m_psFile = new TPostScript(m_psFileName,112);
      m_psFile->Off();
      fStatusBar->SetText("FILE: " + m_psFileName + " is open.");
    }
    }
    break;

  case M_PS_CLOSE:
    {
    if(m_psFile == 0){
      fStatusBar->SetText("NO OPEN PS FILE EXIST!");
    } else{
      m_psFile->On();
      m_psFile->Close();
      fStatusBar->SetText("FILE: " + m_psFileName + " is closed.");
      m_psFile = 0;
    }
    }
    break;

  case M_PS_PRINT:
    {
    if(m_psFileName == ""){
      fStatusBar->SetText("NO SAVED PS FILE EXIST!");
      break;
    } else{
      if(m_psFile != 0){ 
	m_psFile->On(); 
	m_psFile->Close(); 
	m_psFile = 0; 
      } 
      // CERN Bat40 1d cor
      TString printCommand = "xprint -h -p40-1d-cor" + m_psFileName;
      gSystem->Exec(printCommand);
      fStatusBar->SetText("FILE: " + m_psFileName + " is sent to PRINTER: 40-1d-cor");
    }
    }
    break;
   
  case M_PS_DELETE:
    {
    if(m_psFileName == ""){
      fStatusBar->SetText("NO SAVED PS FILE EXIST!");
    } else{
      gSystem->Exec("rm -f " + m_psFileName);
      fStatusBar->SetText("FILE: " + m_psFileName + " is removed.");
    }
    }
    break;

  case M_XMON_ENT1:
    break;

  case M_XMON_ENT2:
    break;
  }
}

//#####################################################################
//#####################################################################
// Launch GUI
void RootUI()
{ 
  gROOT->SetStyle("Plain");
  new XMon(gClient->GetRoot(),200,200);
}
//#####################################################################