/***************************************************************************
 *  Copyright:http://www.naradabrokering.org/index.html
 *  Author: Jaliya Ekanayake
 *  Email : jaliyae@gmail.com
 ****************************************************************************/
#include "NBCCoverMessage.h"
#include "Utils.h"
#include "constants.h"
#include "AgentDiscoveryRequest.h"
#include "AgentDiscoveryResponse.h"
#include "ServiceAdvertiseMessage.h"
#include "ServiceDiscoveryRequest.h"
#include "ServiceDiscoveryResponse.h"

NBCCoverMessage::NBCCoverMessage(): hasExchangeType(false),hasActionType(false), hasActionId(false), hasCorrelationId(false),hasCommTopic(false),hasPayloadMessage(false),deletable(false){}

NBCCoverMessage::~NBCCoverMessage(){
	/*
	 * The only deletable object in this class is the 
	 * payload message. However, the paylaod message is the one
	 * actually needed for the client. So we assume that it is 
	 * client's responsibility to delete the payload message.
	 * */

}

int 
NBCCoverMessage::getSerializedLength(){
	//cout<<"Size of NBCCoverMessage "<<NUM_OF_BYTES_FOR_MESSAGE_HEADERS<<" + "<<payloadMessage->getSerializedLength()<<endl;
	return NUM_OF_BYTES_FOR_MESSAGE_HEADERS + payloadMessage->getSerializedLength();
}	

bool 
NBCCoverMessage::fromBytes(char *bytes){
	 long long offset =0;
 	  	
 	/*Validate the first two bytes*/
 	if(CLARENS_DATA!=Utils::readByte(bytes+offset,offset)) {return false;} ;
	if(NB_CLARENS_MESSAGE!=Utils::readByte(bytes+offset,offset)) {return false;} ;
	
	this->dataType= Utils::readByte(bytes+offset,offset);  
	
	/*Read exchange type*/
	hasExchangeType =Utils::readBool(bytes+offset,offset); 
	if (hasExchangeType) {
		this->exchangeType=Utils::readByte(bytes+offset,offset);
	}
	
	/*Read Action Type*/
	hasActionType =Utils::readBool(bytes+offset,offset); 
	if (hasActionType) {
		this->actionType=Utils::readByte(bytes+offset,offset);
	}
	
	/*Read actionId*/
	this->hasActionId =Utils::readBool(bytes+offset,offset); 
	if(hasActionId){
		int actionIdLength = Utils::readInt(bytes+offset,offset);
		
		char _actionId[actionIdLength];
		Utils::copyBytes((bytes+offset),_actionId,actionIdLength,offset);
		this->actionId.assign(_actionId,actionIdLength);		 
	}
	
	
	/*Read correlationId*/
	this->hasCorrelationId =Utils::readBool(bytes+offset,offset); 
	if(hasCorrelationId){
		int correlationIdLength = Utils::readInt(bytes+offset,offset);
		
		char _corrleationId[correlationIdLength];
		Utils::copyBytes((bytes+offset),_corrleationId,correlationIdLength,offset);
		this->correlationId.assign(_corrleationId,correlationIdLength);		
	}
	
	
	/*Read Communication Topic*/
	this->hasCommTopic =Utils::readBool(bytes+offset,offset); 
	if(hasCommTopic){
		int commTopicLength = Utils::readInt(bytes+offset,offset);
		
		char _commTopic[commTopicLength];
		Utils::copyBytes((bytes+offset),_commTopic,commTopicLength,offset);
		this->commTopic.assign(_commTopic,commTopicLength);
		
	}
	
	/*Read payload*/
	this->hasPayloadMessage =Utils::readBool(bytes+offset,offset); 
	if(hasPayloadMessage){
		int payloadLength = Utils::readInt(bytes+offset,offset);

		
		char _payload[payloadLength];
		Utils::copyBytes((bytes+offset),_payload,payloadLength,offset);		
		 
		/* Message type checking can be refactored */
		if ((this->exchangeType == AGENT_DISCOVERY)	&& (this->actionType ==  REQUEST)) {
			payloadMessage = new AgentDiscoveryRequest();
			payloadMessage->fromBytes(_payload);
			this->deletable=true;
		} else if ((this->exchangeType ==  ROOTLET_DISCOVERY)&& (this->actionType ==  REQUEST)) {
			payloadMessage = new ServiceDiscoveryRequest();
			payloadMessage->fromBytes(_payload);
		} else if ((this->exchangeType ==  ROOTLET_DISCOVERY)&& (this->actionType ==  RESPONSE)) {			
			payloadMessage = new ServiceDiscoveryResponse();
			payloadMessage->fromBytes(_payload);
		} else if ((this->exchangeType ==  AGENT_DISCOVERY)&& (this->actionType ==  RESPONSE)) {			
			payloadMessage = new AgentDiscoveryResponse();
			payloadMessage->fromBytes(_payload);
			this->deletable=true;		
		}else if ((this->exchangeType ==  ROOTLET_ADVERTISE) && (this->actionType ==  REQUEST)) {
			payloadMessage = new ServiceAdvertiseMessage();
			payloadMessage->fromBytes(_payload);
			this->deletable=true;
		}		 
	}
	return true;
}
			
/** Get a byte stream representation of the event */
void
NBCCoverMessage::getBytes(char *buff,int *length){
	long long offset = 0;
	/*Headerse*/
	Utils::writeByte(buff+offset,(char)CLARENS_DATA,offset);
    Utils::writeByte(buff+offset,(char)NB_CLARENS_MESSAGE,offset);
     
    /*DataType*/
    Utils::writeByte(buff+offset,(char)dataType,offset);     
     
    /*Exchange type */
	Utils::writeBool(buff+offset,hasExchangeType,offset);
    if (hasExchangeType) {
		Utils::writeByte(buff+offset,(char) exchangeType,offset);
	}

	/* Action type */
	Utils::writeBool(buff+offset,hasActionType,offset);
	if (hasActionType) {
 		Utils::writeByte(buff+offset,actionType,offset);
    }
    
    /*ActionId*/ 
    Utils::writeBool(buff+offset,hasActionId,offset);
	if (hasActionId) {
		Utils::writeInt(buff+offset,actionId.size(),offset);
		Utils::writeBytes(buff+offset,actionId.c_str(),actionId.size(),offset);
	} 
    
    /*CorrelationId*/ 
    Utils::writeBool(buff+offset,hasCorrelationId,offset);
	if (hasCorrelationId) {
		Utils::writeInt(buff+offset,correlationId.size(),offset);
		Utils::writeBytes(buff+offset,correlationId.c_str(),correlationId.size(),offset);
	} 
	
	/*Communication Topic*/
    Utils::writeBool(buff+offset,hasCommTopic,offset);
	if (hasCommTopic) {
		Utils::writeInt(buff+offset,commTopic.size(),offset);
		Utils::writeBytes(buff+offset,commTopic.c_str(),commTopic.size(),offset);
	} 
	
	/*Payload Message*/
    Utils::writeBool(buff+offset,hasPayloadMessage,offset);
	if (hasPayloadMessage) {
		//Get the length of the message (this is an estimate)
		int estimated_length_of_msg= payloadMessage->getSerializedLength();
		int correct_length_of_msg=0;
		//Create a buffer
		char bytes[estimated_length_of_msg];
		//Get the bytes and update the length
		payloadMessage->getBytes(bytes,&correct_length_of_msg);
		
		if(estimated_length_of_msg<correct_length_of_msg){
			cout<<"ERROR : message size error @ NBCCoverMessage"<<endl;
			cout<<"estimated length = "<<estimated_length_of_msg<<"correct length ="<<correct_length_of_msg<<endl;
			exit(1);
		}
		
		//Now write the length
		//cout<<"Payload lenght @ NBCCoverMessage::getBytes = "<<correct_length_of_msg<<endl;
		Utils::writeInt(buff+offset,correct_length_of_msg,offset);
		//Next the bytes
		Utils::writeBytes(buff+offset,bytes,correct_length_of_msg,offset);
	} 
     
    //Now set the length
    (*length)=offset;          
}
