/***************************************************************************
 *  Copyright:http://www.naradabrokering.org/index.html
 *  Author: Jaliya Ekanayake
 *  Email : jaliyae@gmail.com
 ****************************************************************************/
#include <iostream>
#include <string>
#include "TCServerUIModel.h"
#include "TCanvas.h"
#include <sstream>

ClassImp(TCServerUIModel)

TCServerUIModel::TCServerUIModel(TCanvas *canv,double hGap, double y):canvas(canv),columnNum(-1),horizGap(hGap),topY(y),numFiles(0),numFilesPerRootlet(0),lblPaveText(NULL)
{}

TCServerUIModel::~TCServerUIModel() {
	map<string,TPaveText*>::iterator iter;
	for (iter = rootlets.begin(); iter != rootlets.end(); iter++) {
		delete iter->second;
	}
	rootlets.clear();
	if (lblPaveText!=NULL) {
		delete lblPaveText;
	}
}

void TCServerUIModel::drawHeader() {
	double lblTopX=(this->columnNum)*colWith+horizGap*this->columnNum+horizGap;
	if (lblPaveText==NULL) {
		TPaveText *lblPText = new TPaveText(lblTopX, topY,lblTopX+colWith,topY-headerHeight,"brNDC");
		lblPText->AddText(this->serverName.c_str());

		stringstream ss;
		ss<<this->numFiles;
		ss<<" files";
		lblPText->AddText(ss.str().c_str());
		string available="available";
		lblPText->AddText(available.c_str());
		lblPText->SetFillColor(8);
		lblPText->SetTextAlign(22);
		lblPText->SetBorderSize(1);
		this->lblPaveText=lblPText;
		lblPText->Draw();

	} else {
		this->lblPaveText->SetX1NDC(lblTopX);
		this->lblPaveText->SetX2NDC(lblTopX+colWith);
		this->lblPaveText->Draw();
	}
}
void TCServerUIModel::drawRootlets() {
	if (numFilesPerRootlet>0) {
		if (this->rootlets.size()>0) {
			map<string,TPaveText*>::iterator iter;
			for (iter = rootlets.begin(); iter != rootlets.end(); iter++) {
				iter->second->Draw();
			}
		} else {

			//Now the header is drawn and the number of files per rootlet is defined.
			//cout<<"Drawing rootlets "<<numFilesPerRootlet<<endl;
			int num_rootlets=this->numFiles/numFilesPerRootlet;
			if (this->numFiles%numFilesPerRootlet!=0) {
				num_rootlets++;
			}
			
			double rootlet_height=rHeight/num_rootlets;
			
			for (int i=0; i<num_rootlets; i++) {
				double rTopX=(this->columnNum)*colWith+horizGap*this->columnNum+horizGap;
				double rTopY=this->topY-this->headerHeight-gapBtnServerAndR-(i*rootlet_height);

				TPaveText *rText = new TPaveText(rTopX, rTopY,rTopX+colWith,rTopY-rootlet_height,"brNDC");
				//rText->AddText(this->serverName.c_str());
				string rName=getRootletName(i);
				
				rText->AddText(rName.c_str());
				rText->SetFillColor(30);
				rText->SetTextAlign(12);
				rText->SetBorderSize(1);
				this->rootlets.insert(make_pair(rName, rText));
				rText->Draw();
				//cout
				//		<<"***********************************************ROOOTLET NAME= "
				//		<<rName<<" X1 Y1 X2 Y2 "<<rTopX<<" "<<rTopY<<" "
				//		<<rTopX +colWith<<"  "<<rTopY-rootlet_height<<endl;
			}
		}
	}
}

string TCServerUIModel::getRootletName(int index){
	stringstream ss;
	ss<<this->serverName;
	ss<<"_R_";
	ss<<index;
	
	return ss.str();
}

void TCServerUIModel::draw(int col) {
	this->canvas->cd();
	//Have to draw everything.
	this->columnNum=col;
	drawHeader();
	drawRootlets();
}

void TCServerUIModel::updateRootlet(int num,int status) {
	this->canvas->cd();	
	string rName=getRootletName(num);
	map<string,TPaveText*>::iterator iter= rootlets.find(rName);
	if( iter != rootlets.end() ) {
			if(status==BEFORE_MERGE){
			 iter->second->	SetFillColor(44);
			}else if(status==AFTER_MERGE){
				iter->second->	SetFillColor(38);
			}
			iter->second->Draw();			
	}
}

