/***************************************************************************
 *  Copyright:http://www.naradabrokering.org/index.html
 *  Author: Jaliya Ekanayake
 *  Email : jaliyae@gmail.com
 ****************************************************************************/
#include <iostream>
#include <string>
#include <sstream>

#include "TFileCanvasManager.h"
#include "TCanvas.h"

ClassImp(TFileCanvasManager)

TFileCanvasManager::TFileCanvasManager(TCanvas *canv):canvas(canv),hasUpdates(false),nextFreeColumn(0),numFilesPerRootlet(0){}

TFileCanvasManager::~TFileCanvasManager(){
	map<string,TCServerUIModel*>::iterator iter;
	for(iter = servers.begin(); iter != servers.end(); iter++)
	{
	   	delete iter->second;
	}
	servers.clear();	
}

void 
TFileCanvasManager::setNumFilePerRootlet(int num){
	this->numFilesPerRootlet=num;
	map<string,TCServerUIModel*>::iterator iter;
	for(iter = servers.begin(); iter != servers.end(); iter++)
	{
	   	iter->second->setNumFilesPerRootlet(num);
	}
	this->hasUpdates=true;		
}

void 
TFileCanvasManager::addServer(string url, string name, int numFiles){
		
	double lblTopY=1-lblTopGap;//(0.0,0.0) means bottom left corner		
	TCServerUIModel *server=new TCServerUIModel(this->canvas,lblHorizontalGap,lblTopY);
	server->setServerURL(url);
	server->setServerName(name);
	server->setNumFiles(numFiles);
	server->setNumFilesPerRootlet(this->numFilesPerRootlet);
	this->servers.insert(make_pair(url, server));
	nextFreeColumn++;
	this->hasUpdates=true;
}

void 
TFileCanvasManager::removeServer(string url){
	  map<string,TCServerUIModel*>::iterator iter= servers.find(url);
	  if( iter != servers.end() ) {
		delete iter->second;
	    this->servers.erase(iter);
	  }	
}

TCServerUIModel * 
TFileCanvasManager::getTCServerUIModel(string url){
	  map<string,TCServerUIModel*>::iterator iter= servers.find(url);
	  if( iter != servers.end() ) {
		return iter->second;	    
	  }else{
	  	return NULL;
	  }
}

void 
TFileCanvasManager::updateRootlet(string url,int num, int status){
	 map<string,TCServerUIModel*>::iterator iter= servers.find(url);
	 if( iter != servers.end() ) {
		 iter->second->updateRootlet(num,status);	    
	 }
	 this->canvas->Update();
}

void 
TFileCanvasManager::updateCanvas(){
	if(this->hasUpdates){
		cout<<"TFileCanvasManager::updating"<<endl;
				
		this->canvas->cd();
		TCServerUIModel *uiModel;
		map<string,TCServerUIModel*>::iterator iter;
		int count=0;
		for(iter = servers.begin(); iter != servers.end(); iter++)
		{		
		   	uiModel=iter->second;
		   	uiModel->draw(count);
		   	count++;
		}
		this->canvas->Update();	
		this->hasUpdates=false;
	}
}

void 
TFileCanvasManager::draw(){
	if(this->hasUpdates){
		cout<<"TFileCanvasManager::drawing"<<endl;				
		this->canvas->cd();
		this->canvas->Clear();
		TCServerUIModel *uiModel;
		map<string,TCServerUIModel*>::iterator iter;
		int count=0;
		for(iter = servers.begin(); iter != servers.end(); iter++)
		{
		   	uiModel=iter->second;
		   	uiModel->draw(count); 
			count++;
		}
		this->canvas->Update();
		this->hasUpdates=false;
	}	
	
	/*
	//TODO: Extract the constants.
		rootFileCanvas->cd();
		double borderGapUp=0.02;
		double borderGapDown=0.08;
		double lblGap=0.06;

		double paveAreaHeight=1-borderGapUp-borderGapDown-lblGap;
		double paveAreaTop=paveAreaHeight+borderGapDown;
		double gap=0.005;

		double height=paveAreaHeight/((double)num_files);
		height=height-gap;
		double heightWithGap=height+gap;

		double xVal1=0.05+(currentServerPosInCanvas-1)*0.25;
		double xVal2=xVal1+0.20;

		double lblHeight=0.04;
		double lblTop=1- borderGapUp;
		double lblBottom=lblTop-lblHeight;



		TPaveText *lblPaveText = new TPaveText(xVal1, lblTop,xVal2,lblBottom,"brNDC");
		lblPaveText->AddText(lblStr.c_str());
		lblPaveText->SetFillColor(8);
		lblPaveText->SetTextAlign(12);
		lblPaveText->SetBorderSize(1);
		lblPaveText->Draw();
		clarensServer->setLabelPaveText(lblPaveText);

		for (int i=0; i<num_files; i++) {
			TFileInfo *fi = (TFileInfo *) l->At(i);
			double y1=paveAreaTop-(((double)i)*heightWithGap);
			double y2=y1-height;
			//cout<<" X 1= "<<y1<<"  X2= "<<y2<<endl;

			TPaveText *pt = new TPaveText(xVal1, y1,xVal2,y2,"brNDC");
			pt->SetFillColor(18);
			pt->SetTextAlign(12);
			pt->SetBorderSize(1);

			pt->AddText(fi->name);
			pt->Draw();		
			
			string fileName="/home/clarens/rootfiles/";
			fileName.append(fi->name);
			clarensServer->addRootFile(fileName);
			
		}
		rootFileCanvas->Update();
		return true;
		
		*/
}	

