//Class UserRSA

//    Copyright (C) 2002, 2003 California Institute of Technology
//    Author: Conrad D. Steenberg <conrad@hep.caltech.edu>
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

/*********************************************************************
 * TODO:This is a copy of the UserRSA from Clarens project.
 * Remove this class and change the make file to use the 
 * Clarens header files instad. * 
 * ********************************************************************/


# include "UserRSA.h"
# include <fstream>
//# include <qregexp.h>
//# include <qinputdialog.h> 
# include "openssl/x509.h"
# include "openssl/pem.h"
# include "openssl/bio.h"
# include "openssl/err.h"
# include <sys/types.h>

#include <iostream>



                             

UserRSA*  UserRSA::m_self = 0;

//<<<<<< PUBLIC VARIABLE DEFINITIONS                                    >>>>>>
//<<<<<< CLASS STRUCTURE INITIALIZATION                                 >>>>>>
//<<<<<< PRIVATE FUNCTION DEFINITIONS                                   >>>>>>
//<<<<<< PUBLIC FUNCTION DEFINITIONS                                    >>>>>>
//<<<<<< MEMBER FUNCTION DEFINITIONS                                    >>>>>>

UserRSA*
UserRSA::getUserRSA (const string& cert, const string& key)
{
    if (!m_self)
      m_self = new UserRSA (cert, key);
    return m_self;
}

UserRSA*
UserRSA::getUserRSA (const string& proxy)
{
    if (!m_self)
      m_self = new UserRSA (proxy);
    return m_self;
}

UserRSA*
UserRSA::getUserRSA ()
{ return getUserRSA ("", ""); }

void
UserRSA::ref ()
{ m_ref++; }

void
UserRSA::unref ()
{
    m_ref--;
    if (m_ref <= 0)
    {
      delete m_self;
      m_self = 0;
    }
}

UserRSA::UserRSA   (const string& cert,
                        const string& key)
    : m_userPrivateRSA (0),
      m_userPublicRSA  (0),
      m_cert           (""),
      m_userDN         (""),
      m_error          (""),
      m_initialized    (false),
      m_ref            (0)
{ 
    SSLInit ();
    m_initialized = init (cert, key);
}

UserRSA::UserRSA (const string& proxy)
   : m_userPrivateRSA (0),
      m_userPublicRSA  (0),
      m_cert           (""),
      m_userDN         (""),
      m_error          (""),
      m_initialized    (false),
      m_ref            (0)
{
     SSLInit ();
     m_initialized = init (proxy);     
      
}

UserRSA::~UserRSA ()
{  
    if (m_userPrivateRSA)
      RSA_free (m_userPrivateRSA);
    if (m_userPublicRSA)
      RSA_free (m_userPublicRSA);
}

bool UserRSA::init(string proxy){
   
   cout<<" Loading proxy certificate ..."+proxy<<endl;
   BIO* bio = NULL;
    X509* user_cert;
    m_error = "";
    m_initialized = false;
    
  //  cout<<" Initializing UserRSA ...."<<endl;
    
    //Load the user's public certificate,
    //try the globus-proxy-init version first
    string cert ="";
    string key  ="";
    
    if (cert == "" )
    {
        cert=proxy;
    }
    if ( key == "")
    {
        key=proxy;
    }
    
    if ( key == "")
    {
        m_error = 
	  "Can't find user cert/key pair files or the X509_USER_PROXY file.";
        return m_initialized;
    }

    //Open the user certificate file
    bio = BIO_new_file(cert.c_str (),"r");
    if (!bio)
    {        
        m_error = string ("Fail reading certificate file \"")
	          + cert
		  + "\".";
        return m_initialized;
    }

    //read the user certificate file
    user_cert = PEM_read_bio_X509 (bio, NULL,NULL,NULL);
    BIO_free (bio);
    if (!user_cert)
    {
        m_error = string ("Fail reading user public key from file \"")
	          + cert 
		  + "\".";
        return m_initialized;
    }

    //Get the public key from user certificate
    EVP_PKEY * user_pub_key = X509_get_pubkey (user_cert);
    X509_free (user_cert);
    if (!user_pub_key)
    { 
        unsigned long err = ERR_get_error ();
        m_error = string ("Could not read user public key:\n")
                  + ERR_reason_error_string (err);
        return m_initialized;
    }
    
    //Get the user Public RSA from the public key
    m_userPublicRSA = EVP_PKEY_get1_RSA (user_pub_key);
    EVP_PKEY_free (user_pub_key);
    if (!m_userPublicRSA)
    { 
        unsigned long err = ERR_get_error ();
        m_error =
	  string ("Could not extract RSA structure from user public key:\n")
          + ERR_reason_error_string (err);
        return m_initialized;
    }

    // Open the user's private key file
    bio = BIO_new_file(key.c_str(),"r");
    if (!bio)
    {
        RSA_free (m_userPublicRSA);
        m_userPublicRSA = 0;
        m_error = string ("Fail reading user private key file: \"")
	          + key
		  + "\".";
        return m_initialized;
    }
    
    //Parse the User Certificate file and extract User DN and certificate
    char buf[256];
    bool is_cert = false;
    std::ifstream fin;
    fin.open (cert.c_str ());
    char cert_buf[160];
    bool isCert=false;
    
    m_cert="";
    while(!fin.eof()){
       fin.getline(buf,160);
       string cert_line(buf);
       
       if((cert_line.find("-----BEGIN CERTIFICATE-----",0) !=string::npos) || 
           (cert_line.find("-----BEGIN X509 CERTIFICATE-----",0) !=string::npos)){
	      cout<<" reading certificate "<<endl;
	      isCert=true;	   
       }
       
       if(isCert)
          m_cert += cert_line +"\n";
       
       if((cert_line.find("-----END CERTIFICATE-----",0) !=string::npos) || 
           (cert_line.find("-----END X509 CERTIFICATE-----",0) !=string::npos)){
	     // cout<<" reading certificate finsihed"<<endl;
	     // cout<<m_cert<<endl;
	      isCert=false;	   
       }
    
    }
    
    fin.close ();
    
    //Read the user Private key, and if pass phrase needed the call back the
    //passPhraseInput member function.    
    m_userPrivateRSA = PEM_read_bio_RSAPrivateKey (bio,
                                                   NULL,
						   passPhraseInput,
						   (void*)this);
    
   // cout<<" Initialized Private RSA"<<m_userPrivateRSA<<endl;
    BIO_free (bio);
    if (!m_userPrivateRSA)
    { 
        unsigned long err = ERR_get_error ();
        m_error =
	  string ("Could not extract RSA from user private key:\n")
                  + ERR_reason_error_string (err);
        cout<<m_error<<endl;
	RSA_free (m_userPublicRSA);
        m_userPublicRSA = 0;
        return m_initialized;
    }

    m_initialized = true;
    return m_initialized;

}

bool 
UserRSA::init (string cert, string key)
{
    BIO* bio = NULL;
    X509* user_cert;
    m_error = "";
    m_initialized = false;
    
  //  cout<<" Initializing UserRSA ...."<<endl;
    
    //Load the user's public certificate,
    //try the globus-proxy-init version first
    if (cert == "" )
    {
        if (getenv ("X509_USER_PROXY") != NULL)
          cert = getenv ("X509_USER_PROXY");
        else if (getenv ("HOME") != NULL){
          cert = string (getenv ("HOME")) + "/.globus/usercert.pem";
	  cout<<" Reading user cert from "+string (getenv ("HOME")) + "/.globus/usercert.pem"<<endl;
	
	}
    }
    if ( key == "")
    {
        if (getenv ("X509_USER_PROXY") != NULL)
          key = getenv ("X509_USER_PROXY");
        else if (getenv ("HOME") != NULL){
          key = string (getenv ("HOME")) + "/.globus/userkey.pem";
	  cout<<" Reading user Key from "+string (getenv ("HOME")) + "/.globus/userkey.pem"<<endl;
	}
    }
    
    if ( key == "")
    {
        m_error = 
	  "Can't find user cert/key pair files or the X509_USER_PROXY file.";
        return m_initialized;
    }

    //Open the user certificate file
    bio = BIO_new_file(cert.c_str (),"r");
    if (!bio)
    {        
        m_error = string ("Fail reading certificate file \"")
	          + cert
		  + "\".";
        return m_initialized;
    }

    //read the user certificate file
    user_cert = PEM_read_bio_X509 (bio, NULL,NULL,NULL);
    BIO_free (bio);
    if (!user_cert)
    {
        m_error = string ("Fail reading user public key from file \"")
	          + cert 
		  + "\".";
        return m_initialized;
    }

    //Get the public key from user certificate
    EVP_PKEY * user_pub_key = X509_get_pubkey (user_cert);
    X509_free (user_cert);
    if (!user_pub_key)
    { 
        unsigned long err = ERR_get_error ();
        m_error = string ("Could not read user public key:\n")
                  + ERR_reason_error_string (err);
        return m_initialized;
    }
    
    //Get the user Public RSA from the public key
    m_userPublicRSA = EVP_PKEY_get1_RSA (user_pub_key);
    EVP_PKEY_free (user_pub_key);
    if (!m_userPublicRSA)
    { 
        unsigned long err = ERR_get_error ();
        m_error =
	  string ("Could not extract RSA structure from user public key:\n")
          + ERR_reason_error_string (err);
        return m_initialized;
    }

    // Open the user's private key file
    bio = BIO_new_file(key.c_str(),"r");
    if (!bio)
    {
        RSA_free (m_userPublicRSA);
        m_userPublicRSA = 0;
        m_error = string ("Fail reading user private key file: \"")
	          + key
		  + "\".";
        return m_initialized;
    }
    
    //Parse the User Certificate file and extract User DN and certificate
    char buf[256];
    bool is_cert = false;
    std::ifstream fin;
    fin.open (cert.c_str ());
    char cert_buf[160];
    bool isCert=false;
    
    m_cert="";
    while(!fin.eof()){
       fin.getline(buf,160);
       string cert_line(buf);
       
       if((cert_line.find("-----BEGIN CERTIFICATE-----",0) !=string::npos) || 
           (cert_line.find("-----BEGIN X509 CERTIFICATE-----",0) !=string::npos)){
	      cout<<" reading certificate "<<endl;
	      isCert=true;	   
       }
       
       if(isCert)
          m_cert += cert_line +"\n";
       
       if((cert_line.find("-----END CERTIFICATE-----",0) !=string::npos) || 
           (cert_line.find("-----END X509 CERTIFICATE-----",0) !=string::npos)){
	     // cout<<" reading certificate finsihed"<<endl;
	     // cout<<m_cert<<endl;
	      isCert=false;	   
       }
    
    }
    
    fin.close ();
    
    //Read the user Private key, and if pass phrase needed the call back the
    //passPhraseInput member function.    
    m_userPrivateRSA = PEM_read_bio_RSAPrivateKey (bio,
                                                   NULL,
						   passPhraseInput,
						   (void*)this);
    
   // cout<<" Initialized Private RSA"<<m_userPrivateRSA<<endl;
    BIO_free (bio);
    if (!m_userPrivateRSA)
    { 
        unsigned long err = ERR_get_error ();
        m_error =
	  string ("Could not extract RSA from user private key:\n")
                  + ERR_reason_error_string (err);
        cout<<m_error<<endl;
	RSA_free (m_userPublicRSA);
        m_userPublicRSA = 0;
        return m_initialized;
    }

    m_initialized = true;
    return m_initialized;
}

const RSA*
UserRSA::getPrivateRSA ()const
{ return m_userPrivateRSA; }

const RSA*
UserRSA::getPublicRSA () const
{ return m_userPublicRSA; }




 string UserRSA::getUserCert ()  
{ 
   
   //cout<<" Returning the user cert "<<m_cert<<endl;
   return m_cert;
   
   }

 
 
 char* UserRSA::getUserDN ()  
{ 

  int size=m_userDN.length();
  char buf[size+1];
  // for(int i=0;i<size;i++)
  //   buf[i]=m_userDN[i];
  const char* buf1 =m_userDN.c_str();
  for (int i=0;i==size;i++)
    buf[i]=buf1[i];
   
   char *bufA =buf;
   return bufA;
   }

const string&
UserRSA::error () const
{ return m_error; }

bool
UserRSA::isInitialized () const
{ return m_initialized; }

void 
UserRSA::SSLInit ()
{
    OpenSSL_add_all_algorithms ();
    OpenSSL_add_all_ciphers ();
    OpenSSL_add_all_digests ();
    ERR_load_crypto_strings ();
}

int
UserRSA::passPhraseInput (char *buf, int size, int /*rwflag*/, void *object)
{
    
    string passwd;
    cout<<"**********************************************"<<endl;
    cout<<"*****  Please Input Key password   ***********"<<endl;
    cout<<"***********************************************"<<endl;
    cin>>passwd;
    //passwd ="33596a";
    size = passwd.length();
   
    
    
    for(int i=0;i<size;i++)
       buf[i]=passwd[i];
    //buf=cert_buf;
    
   // cout<<" Returning password"<<buf<<endl;
    return size;
}
